from collections import namedtuple
from decimal import Decimal
from enum import Enum
from functools import partial
from typing import Any, Dict, List, Union
from uuid import uuid4

import motor.motor_asyncio
import pymongo
import pytz
from bson import CodecOptions, Decimal128, ObjectId
from bson.codec_options import TypeRegistry
from bson.json_util import dumps, loads
from gridfs import GridFS, GridFSBucket, GridOut
from jcramda import (
    _,
    attr,
    compose,
    curry,
    enum_name,
    first,
    getitem,
    has_attr,
    if_else,
    in_,
    is_a,
    locnow,
    not_,
    obj,
    popitem,
    when,
)
from pymongo.collection import Collection, ReturnDocument
from pymongo.database import Database
from pymongo.results import InsertOneResult, UpdateResult

from jcutil.core import get_running_loop


def fallback_encoder(value):
    return when(
        (is_a(Decimal), Decimal128),
        # (is_a(NDFrame), df_to_dict),
        (is_a(InsertOneResult), compose(obj('insertedId'), attr('inserted_id'))),
        (is_a(UpdateResult), compose(obj('upsertedId'), attr('upserted_id'))),
        (has_attr('get'), lambda o: o.get()),
        (has_attr('result'), lambda o: o.result()),
        else_=str
    )(value)


# class NDFrameCodec(TypeCodec):
#     python_type = pd.Series
#     bson_type = dict
#
#     def transform_bson(self, value):
#         return pd.Series(value)
#
#     def transform_python(self, value):
#         return df_to_dict(value)


_type_registry = TypeRegistry(
    # type_codecs=[NDFrameCodec()],
    fallback_encoder=fallback_encoder
)


class UniqFileGridFSBucket(GridFSBucket):
    def _find_one(self, filename):
        return [*self.find({'filename': filename}).sort('uploadDate', -1).limit(1)]

    def _create_proxy(self, out: GridOut, opened=False):
        fid: ObjectId = getattr(out, '_id')
        self.delete(fid)
        create_method = self.open_upload_stream_with_id if opened \
            else self.upload_from_stream_with_id
        return partial(create_method, fid)

    def open_save_file(self, filename, **kwargs):
        open_by_id = compose(
            lambda grid_out: self._create_proxy(
                grid_out, True)(filename, **kwargs),
            first,
        )
        return compose(
            if_else(not_,
                    lambda _: self.open_upload_stream(filename, **kwargs),
                    open_by_id),
            self._find_one,
        )(filename)

    def save_file(self, filename, **kwargs):
        upload_by_id = compose(
            lambda grid_out: self._create_proxy(grid_out)(filename, **kwargs),
            first,
        )

        return compose(
            if_else(not_,
                    lambda _: self.upload_from_stream(filename, **kwargs),
                    upload_by_id),
            self._find_one,
        )(filename)


class AsyncUniqFileGridFSBucket:
    """异步版本的UniqFileGridFSBucket"""
    def __init__(self, db, bucket_name="fs"):
        self.bucket = motor.motor_asyncio.AsyncIOMotorGridFSBucket(db, bucket_name=bucket_name)

    async def _find_one(self, filename):
        cursor = self.bucket.find({'filename': filename}).sort('uploadDate', -1).limit(1)
        docs = []
        async for doc in cursor:
            docs.append(doc)
        return docs

    async def delete(self, file_id):
        await self.bucket.delete(file_id)

    async def open_upload_stream(self, filename, **kwargs):
        return await self.bucket.open_upload_stream(filename, **kwargs)

    async def open_upload_stream_with_id(self, file_id, filename, **kwargs):
        return await self.bucket.open_upload_stream_with_id(file_id, filename, **kwargs)

    async def upload_from_stream(self, filename, source, **kwargs):
        return await self.bucket.upload_from_stream(filename, source, **kwargs)

    async def upload_from_stream_with_id(self, file_id, filename, source, **kwargs):
        return await self.bucket.upload_from_stream_with_id(file_id, filename, source, **kwargs)

    async def _create_proxy(self, grid_out: Any, opened=False):
        fid: ObjectId = getattr(grid_out, '_id')
        await self.delete(fid)
        if opened:
            return partial(self.open_upload_stream_with_id, fid)
        else:
            return partial(self.upload_from_stream_with_id, fid)

    async def open_save_file(self, filename, **kwargs):
        docs = await self._find_one(filename)
        if not docs:
            return await self.open_upload_stream(filename, **kwargs)
        else:
            proxy = await self._create_proxy(docs[0], True)
            return proxy(filename, **kwargs)

    async def save_file(self, filename, **kwargs):
        docs = await self._find_one(filename)
        if not docs:
            return partial(self.upload_from_stream, filename, **kwargs)
        else:
            proxy = await self._create_proxy(docs[0])
            return proxy(filename, **kwargs)


class MongoClient:
    """MongoDB客户端，同时提供同步和异步接口"""

    def __init__(self, uri: str, alias: str = None):
        self.uri = uri
        self.alias = alias if alias is not None else uuid4().hex
        self.sync_client = pymongo.MongoClient(uri)

        # Use the same event loop for all async operations
        loop = get_running_loop()
        self.async_client = motor.motor_asyncio.AsyncIOMotorClient(
            uri,
            io_loop=loop
        )

        self.default_db_name = self.sync_client.get_default_database().name if self.sync_client.get_default_database() is not None else None

        # 配置编解码选项
        self.codec_options = CodecOptions(
            tz_aware=True,
            type_registry=_type_registry,
            tzinfo=pytz.timezone('Asia/Shanghai')
        )

    def get_database(self, db_name=None) -> Database:
        """获取同步数据库对象"""
        if db_name is None:
            return self.sync_client.get_default_database()
        return self.sync_client.get_database(db_name)

    def get_async_database(self, db_name=None):
        """获取异步数据库对象"""
        # Ensure we have a valid async client
        self._ensure_valid_async_client()

        if db_name is None:
            if self.default_db_name:
                return self.async_client[self.default_db_name]
            return self.async_client.get_default_database()
        return self.async_client.get_database(db_name)

    def get_collection(self, collection_name: Union[str, Enum], db_name=None) -> Collection:
        """获取同步集合对象"""
        db = self.get_database(db_name)
        collection_name = enum_name(collection_name)
        return db.get_collection(collection_name, codec_options=self.codec_options)

    def get_async_collection(self, collection_name: Union[str, Enum], db_name=None):
        """获取异步集合对象"""
        # Ensure we have a valid async client
        self._ensure_valid_async_client()

        db = self.get_async_database(db_name)
        collection_name = enum_name(collection_name)
        return db.get_collection(collection_name, codec_options=self.codec_options)

    def _ensure_valid_async_client(self):
        """确保异步客户端使用有效的事件循环"""
        try:
            # Try to get the loop used by the async client
            loop = self.async_client.get_io_loop()
            # Check if the loop is closed
            if loop.is_closed():
                # Create a new async client with a valid loop
                loop = get_running_loop()
                self.async_client = motor.motor_asyncio.AsyncIOMotorClient(
                    self.uri,
                    io_loop=loop
                )
        except (RuntimeError, AttributeError):
            # If we can't get the loop or there's another issue,
            # create a new async client with a valid loop
            loop = get_running_loop()
            self.async_client = motor.motor_asyncio.AsyncIOMotorClient(
                self.uri,
                io_loop=loop
            )

    def get_fs(self, db_name=None) -> GridFS:
        """获取同步GridFS对象"""
        db = self.get_database(db_name)
        return GridFS(db)

    def get_fs_bucket(self, db_name=None, bucket_name='fs') -> UniqFileGridFSBucket:
        """获取同步GridFSBucket对象"""
        db = self.get_database(db_name)
        return UniqFileGridFSBucket(db, bucket_name)

    def get_async_fs_bucket(self, db_name=None, bucket_name='fs') -> AsyncUniqFileGridFSBucket:
        """获取异步GridFSBucket对象"""
        db = self.get_async_database(db_name)
        return AsyncUniqFileGridFSBucket(db, bucket_name)

    # 同步操作方法
    def find(self, collection: Union[Collection, str, Enum], query: dict, db_name=None, **kwargs):
        """同步查询多条记录"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)
        return list(collection.find(query, **kwargs))

    def find_one(self, collection: Union[Collection, str, Enum], query: dict, db_name=None, **kwargs):
        """同步查询单条记录"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)
        return collection.find_one(query, **kwargs)

    def find_by_id(self, collection: Union[Collection, str, Enum], _id: Union[str, ObjectId], db_name=None, **kwargs):
        """同步按ID查询"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)
        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        return collection.find_one({'_id': query_id}, **kwargs)

    def find_in_ids(self, collection: Union[Collection, str, Enum], ids: List[str], db_name=None, **kwargs):
        """同步按多个ID查询"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)
        query = {'_id': {'$in': [ObjectId(n) for n in ids if n is not None]}}
        return list(collection.find(query, **kwargs))

    def find_page(self, collection: Union[Collection, str, Enum], query: dict,
                  page_size=10, page_no=1, sort=None, db_name=None, **kwargs):
        """同步分页查询"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)
        skip = page_size * (page_no - 1)
        if sort is None:
            sort = [('createdTime', pymongo.DESCENDING)]
        query.setdefault('logicDeleted', False)
        cursor = collection.find(query, **kwargs).sort(sort).skip(skip).limit(page_size)
        total = collection.count_documents(query)
        return {
            'content': list(cursor),
            'pageNo': page_no,
            'pageSize': page_size,
            'totalCount': total
        }

    def save(self, collection: Union[Collection, str, Enum], data: dict, db_name=None):
        """同步保存数据（插入或更新）"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)

        if '_id' in data:
            result = collection.update_one(
                {'_id': data['_id']},
                {'$set': data, '$currentDate': {'updateTime': True}, '$inc': {'__v': 1}},
                upsert=True
            )
            if result.upserted_id:
                data['_id'] = result.upserted_id
        else:
            data.setdefault('createTime', locnow())
            result = collection.insert_one(data)
            if result.inserted_id:
                data['_id'] = result.inserted_id

        return data

    def replace(self, collection: Union[Collection, str, Enum], _id: Union[str, ObjectId], data: dict, db_name=None):
        """同步替换数据"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)

        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        old = collection.find_one({'_id': query_id})

        if '_id' in data:
            data.pop('_id')

        data.setdefault('__v', 0)
        data['__v'] += 1
        data['updateTime'] = locnow()

        result = collection.replace_one({'_id': query_id}, data, upsert=True)
        if result.modified_count or result.upserted_id:
            return self.find_by_id(collection, query_id), old

        return None, old

    def delete(self, collection: Union[Collection, str, Enum], _id: Union[str, ObjectId], db_name=None):
        """同步删除数据"""
        if isinstance(collection, (str, Enum)):
            collection = self.get_collection(collection, db_name)

        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        result = collection.delete_one({'_id': query_id})
        return result.deleted_count

    # 异步操作方法
    async def async_find(self, collection: Union[str, Enum], query: dict, db_name=None, **kwargs):
        """异步查询多条记录"""
        async_collection = self.get_async_collection(collection, db_name)
        cursor = async_collection.find(query, **kwargs)
        result = []
        async for doc in cursor:
            result.append(doc)
        return result

    async def async_find_one(self, collection: Union[str, Enum], query: dict, db_name=None, **kwargs):
        """异步查询单条记录"""
        async_collection = self.get_async_collection(collection, db_name)
        return await async_collection.find_one(query, **kwargs)

    async def async_find_by_id(self, collection: Union[str, Enum], _id: Union[str, ObjectId], db_name=None, **kwargs):
        """异步按ID查询"""
        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        return await self.async_find_one(collection, {'_id': query_id}, db_name, **kwargs)

    async def async_find_in_ids(self, collection: Union[str, Enum], ids: List[str], db_name=None, **kwargs):
        """异步按多个ID查询"""
        query = {'_id': {'$in': [ObjectId(n) for n in ids if n is not None]}}
        return await self.async_find(collection, query, db_name, **kwargs)

    async def async_find_page(self, collection: Union[str, Enum], query: dict,
                       page_size=10, page_no=1, sort=None, db_name=None, **kwargs):
        """异步分页查询"""
        async_collection = self.get_async_collection(collection, db_name)
        skip = page_size * (page_no - 1)
        if sort is None:
            sort = [('createdTime', pymongo.DESCENDING)]
        query.setdefault('logicDeleted', False)

        cursor = async_collection.find(query, **kwargs).sort(sort).skip(skip).limit(page_size)
        docs = []
        async for doc in cursor:
            docs.append(doc)

        total = await async_collection.count_documents(query)
        return {
            'content': docs,
            'pageNo': page_no,
            'pageSize': page_size,
            'totalCount': total
        }

    async def async_save(self, collection: Union[str, Enum], data: dict, db_name=None):
        """异步保存数据（插入或更新）"""
        async_collection = self.get_async_collection(collection, db_name)

        if '_id' in data:
            result = await async_collection.update_one(
                {'_id': data['_id']},
                {'$set': data, '$currentDate': {'updateTime': True}, '$inc': {'__v': 1}},
                upsert=True
            )
            if result.upserted_id:
                data['_id'] = result.upserted_id
        else:
            data.setdefault('createTime', locnow())
            result = await async_collection.insert_one(data)
            if result.inserted_id:
                data['_id'] = result.inserted_id

        return data

    async def async_replace(self, collection: Union[str, Enum], _id: Union[str, ObjectId], data: dict, db_name=None):
        """异步替换数据"""
        async_collection = self.get_async_collection(collection, db_name)

        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        old = await async_collection.find_one({'_id': query_id})

        if '_id' in data:
            data.pop('_id')

        data.setdefault('__v', 0)
        data['__v'] += 1
        data['updateTime'] = locnow()

        result = await async_collection.replace_one({'_id': query_id}, data, upsert=True)
        if result.modified_count or result.upserted_id:
            new_doc = await self.async_find_by_id(collection, query_id, db_name)
            return new_doc, old

        return None, old

    async def async_delete(self, collection: Union[str, Enum], _id: Union[str, ObjectId], db_name=None):
        """异步删除数据"""
        async_collection = self.get_async_collection(collection, db_name)

        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        result = await async_collection.delete_one({'_id': query_id})
        return result.deleted_count

    def create_proxy(self, collection_name: str, db_name=None):
        """创建集合代理对象，提供简化的操作方法"""
        proxy_obj = namedtuple(f'{self.alias}_{collection_name}',
                            'all,find,add,update,replace,delete')

        def all(**kwargs):
            return self.find(collection_name, {}, db_name, **kwargs)

        def find(query: dict, **kwargs):
            return self.find(collection_name, query, db_name, **kwargs)

        def add(data):
            data.setdefault('createTime', locnow())
            return self.save(collection_name, data, db_name)

        def update(_id, data, **kwargs):
            current = self.find_by_id(collection_name, _id, db_name)
            if current:
                data_to_update = current.copy()
                data_to_update.update(data)
                data_to_update['_id'] = current['_id']
                result = self.save(collection_name, data_to_update, db_name)
                return result
            return None

        def replace(_id, data, **kwargs):
            new_doc, old = self.replace(collection_name, _id, data, db_name)
            return new_doc

        def delete(_id):
            return self.delete(collection_name, _id, db_name)

        return proxy_obj(all, find, add, update, replace, delete)

    async def create_async_proxy(self, collection_name: str, db_name=None):
        """创建异步集合代理对象，提供简化的异步操作方法"""
        proxy_obj = namedtuple(f'{self.alias}_{collection_name}_async',
                            'all,find,add,update,replace,delete')

        async def all(**kwargs):
            return await self.async_find(collection_name, {}, db_name, **kwargs)

        async def find(query: dict, **kwargs):
            return await self.async_find(collection_name, query, db_name, **kwargs)

        async def add(data):
            data.setdefault('createTime', locnow())
            return await self.async_save(collection_name, data, db_name)

        async def update(_id, data, **kwargs):
            current = await self.async_find_by_id(collection_name, _id, db_name)
            if current:
                data_to_update = current.copy()
                data_to_update.update(data)
                data_to_update['_id'] = current['_id']
                result = await self.async_save(collection_name, data_to_update, db_name)
                return result
            return None

        async def replace(_id, data, **kwargs):
            new_doc, old = await self.async_replace(collection_name, _id, data, db_name)
            return new_doc

        async def delete(_id):
            return await self.async_delete(collection_name, _id, db_name)

        return proxy_obj(all, find, add, update, replace, delete)

    def __str__(self):
        return f"MongoClient(uri='{self.uri}', alias='{self.alias}')"

    def __repr__(self):
        return self.__str__()


# 全局客户端实例管理
__clients: Dict[str, MongoClient] = dict()


def exists(name: str):
    """检查指定名称的客户端是否存在"""
    return name in __clients


def new_client(uri: str, alias: str = None) -> MongoClient:
    """创建新的MongoDB客户端实例"""
    db_name = alias if alias is not None else uuid4().hex
    client = MongoClient(uri, db_name)
    __clients[db_name] = client
    return client


def get_client(alias: str = None) -> MongoClient:
    """获取指定别名的客户端实例"""
    if alias is None:
        alias = first(list(__clients.keys()))
    return __clients[alias]


def load(conf: dict):
    """从配置加载多个MongoDB连接"""
    if conf and len(conf) > 0:
        for key in conf:
            new_client(conf[key], key)
            # print(f'mongodb: [{key}] connected')


def instances():
    """获取所有客户端实例名称"""
    return list(__clients.keys())


# 兼容旧API的辅助函数
def conn(key=None) -> MongoClient:
    """获取客户端连接（兼容旧API）"""
    return get_client(key)


def get_collection(tag, collection: Union[str, Enum], db_name=None):
    """获取集合对象（兼容旧API）"""
    client = get_client(tag)
    return client.get_collection(collection, db_name)


def fs_client(key=None) -> GridFS:
    """获取GridFS对象（兼容旧API）"""
    client = get_client(key)
    return client.get_fs()


def fs_bucket(db_name, bucket_name='fs') -> UniqFileGridFSBucket:
    """获取GridFSBucket对象（兼容旧API）"""
    client = get_client(db_name)
    return client.get_fs_bucket(bucket_name=bucket_name)


# 兼容旧API的数据操作函数
@curry
def find(collection, query):
    """查询函数（兼容旧API）"""
    assert is_a(Collection, collection)
    return tuple(collection.find(query))


@curry
def find_one(collection, query):
    """查询单条记录（兼容旧API）"""
    assert is_a(Collection, collection)
    return collection.find_one(query)


@curry
def find_by(collection, key, value):
    """按字段查询（兼容旧API）"""
    return collection.find(filter={key: value})


@curry
def save(collection, data):
    """保存数据（兼容旧API）"""
    if '_id' in data:
        r = collection.update_one(
            {'_id': data['_id']}, {'$set': data}, upsert=True)
        if hasattr(r, 'upserted_id') and r.upserted_id:
            data['_id'] = r.upserted_id
    else:
        r = collection.insert_one(data)
    if hasattr(r, 'inserted_id'):
        data['_id'] = r.inserted_id
    return data


@curry
def replace_one(collection, data):
    """替换数据（兼容旧API）"""
    assert '_id' in data, 'must include _id field in data.'
    query = {'_id': data['_id']}
    old = collection.find_one(query)
    collection.replace_one(query, data, upsert=True)
    return data, old


def find_page(collection, query, page_size=10, page_no=1, sort=None):
    """分页查询（兼容旧API）"""
    assert is_a(Collection, collection)
    skip = page_size * (page_no - 1)
    if sort is None:
        sort = [('createdTime', pymongo.DESCENDING)]
    query['logicDeleted'] = False
    cursor = collection.find(query).sort(sort).skip(skip).limit(page_size)
    total = collection.count_documents(query)
    return {'content': list(cursor), 'pageNo': page_no, 'pageSize': page_size, 'totalCount': total}


def mdb_proxy(tag, collection_name):
    """创建集合代理（兼容旧API）"""
    client = get_client(tag)
    return client.create_proxy(collection_name)


# 工具函数
to_json = dumps
from_json = loads


class BaseModel:
    """基础模型类"""
    def __init__(self, db_or_client, collection_name):
        if isinstance(db_or_client, MongoClient):
            self.client = db_or_client
            self.collection = db_or_client.get_collection(collection_name)
        else:
            self.client = None
            self.collection = db_or_client.get_collection(collection_name)

    def find_by_id(self, _id: Union[str, ObjectId]):
        query_id = _id if isinstance(_id, ObjectId) else ObjectId(_id)
        return self.collection.find_one(filter={'_id': query_id})

    def find_in_ids(self, ids: List[str], *args):
        query = {'_id': {'$in': [ObjectId(n) for n in ids if n is not None]}}
        return list(self.collection.find(query, *args))

    def _save(self, sepc_field: str = None, **kwargs):
        assert '_id' in kwargs or not not_(sepc_field) and sepc_field in kwargs, \
            f'must had [_id] field or [{sepc_field}] field'
        query = if_else(
            in_(_, '_id'),
            compose(obj('_id'), popitem('_id')),
            compose(obj(sepc_field), getitem(sepc_field))
        )
        return self.collection.find_one_and_update(
            filter=query(kwargs),
            update={'$set': kwargs},
            upsert=True,
            return_document=ReturnDocument.AFTER
        )



# need install kafka-python
import asyncio
import json
import os
from enum import Enum, auto
from typing import Any, Callable, Optional, Protocol, Tuple, Union

from kafka import KafkaAdminClient, KafkaConsumer, KafkaProducer
from kafka.admin import NewTopic

from jcutil.core import async_run

__clients = {}

__cache = {
    'idx': 0
}


class AutoOffsetRest(Enum):
    EARLIEST = auto()
    LATEST = auto()
    NONE = auto()

    @property
    def sv(self):
        return self.name.lower()


class BaseMessage(Protocol):
    topic: str
    key: Optional[str]
    timestamp: int
    offset: int
    headers: list


class MockMessage(BaseMessage):
    def __init__(self, **kwargs):
        self.topic = kwargs.get('topic')
        self.key = kwargs.get('key')
        self.timestamp = kwargs.get('timestamp')
        self.offset = kwargs.get('offset', 0)
        self.headers = kwargs.get('headers', [])


def load(conf):
    for key in conf:
        __clients[key] = conf[key]


def new_client(tag, *servers):
    __clients[tag] = ','.join(servers)


def _convert(raw) -> bytes:
    try:
        s = raw if isinstance(raw, str) else json.dumps(raw, ensure_ascii=False)
    except TypeError:
        s = str(raw)

    return s.encode('utf8')


def send(tag, topic, msg, on_success=None, on_error=None, **kwargs):
    assert tag in __clients, 'not found tag in clients'
    producer = KafkaProducer(bootstrap_servers=__clients[tag])
    f = producer.send(topic, _convert(msg), **kwargs)
    if on_success:
        f.add_callback(on_success)
    if on_error:
        f.add_errback(on_error)
    return f


def _value_des(raw):
    s = raw.decode()
    try:
        r = json.loads(s)
        return json.loads(r) if isinstance(r, str) else r
    except (json.JSONDecodeError, ValueError, TypeError):
        return s


async def subscribe(tag, group_id, / ,
                    topics: Union[Tuple, str] = (),
                    offset_reset: AutoOffsetRest = AutoOffsetRest.LATEST,
                    handler: Callable[[Any, BaseMessage], Any] = None,
                    debug=False,
                    **kwargs
                    ):
    """
    监听指定话题，并非阻塞的运行handler方法处理消息

    Parameters
    ----------
    tag
        要监听的mq别称 （事先已注册的tag）
    group_id
        当前消费者的group_id，通group间不会重复消费
    topics
        话题， 支持 "oneTopic", ("foo", "bar"), "^foo.*" 三种形式
    offset_reset: AutoOffsetRest
        消费方式，默认：'latest'，从消息队列尾端开始消费
    handler: Callable[[Any, Message], Any]
        回调函数，用于处理消息
    debug: bool
        是否开启debug，默认False. 开启后会输出接收到的消息对象
    client_id: str 客户端名称
    value_deserializer: Callable 对值进行反序列化的方法，默认为内置的：_value_des方法

    Returns
    -------


    Examples
    -------------------
        import asyncio
        asyncio.run(subscribe('someTag', 'testGroup', 'testTopic', handler=print))
    """
    assert tag in __clients
    config = dict(
        client_id=kwargs.get('client_id', f'{group_id}-{os.getpid()}-{__cache["idx"]}'),
        bootstrap_servers=__clients[tag],
        group_id=group_id,
        auto_offset_reset=offset_reset.sv,
        value_deserializer=kwargs.get('value_deserializer', _value_des),
    )
    __cache['idx'] += 1
    consumer = KafkaConsumer(**config)
    topics_params = dict(topics=topics) if isinstance(topics, Tuple) \
        else dict(pattern=topics)

    consumer.subscribe(**topics_params)
    print('group_id: {}, subscribe topics: {} '.format(group_id, topics_params))

    for msg in consumer:
        try:
            coro = handler(msg.value, msg) if asyncio.iscoroutinefunction(handler) \
                else async_run(handler, msg.value, msg)
            r = await coro
            while asyncio.isfuture(r):
                r = await r
            if debug:
                if r == 'exit':
                    break
                print(msg, r)
        except RuntimeError as err:
            print(err)
    print('stop subscribe.')


def create_topic(tag, topic):
    admin = KafkaAdminClient(bootstrap_servers=__clients[tag])
    new_topic = NewTopic(topic, 3, 1)
    admin.create_topics((new_topic,))




# pandas tools
from jcramda import attr, compose, curry, flat_concat, is_a, partial

try:
    import numpy as np
    import pandas as pd
    col_value = curry(lambda col, df: compose(lambda s: s.iloc[0], attr(col))(df))

    df_dt = partial(pd.to_datetime, errors='ignore')

    df_to_json = curry(lambda df: df.to_json(
        orient='records',
        default_handler=str,
        force_ascii=False)
    )

    ser_to_json = curry(lambda ser: ser.to_json(
        default_handler=str,
        force_ascii=False,
    ))


    def df_to_dict(df):
        return df.to_dict(orient='records') if isinstance(df, pd.DataFrame) else df.to_dict()


    TYPE_REGS = (
        (is_a(pd.DataFrame), lambda df: flat_concat(df.to_dict(orient='records'))),
        (is_a(pd.Series), lambda ser: flat_concat(ser.to_dict())),
        (is_a(np.bool_), bool),
    )

    __all__ = (
        'col_value',
        'df_dt',
        'df_to_json',
        'ser_to_json',
        'df_to_dict',
        'TYPE_REGS',
    )
except ModuleNotFoundError:
    TYPE_REGS = ()
    __all__ = ('TYPE_REGS',)



import logging
import os
import socket
from functools import lru_cache
from pathlib import Path

from dotenv import load_dotenv


@lru_cache()
def local_ip():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(('180.76.76.76', 80))
        r = s.getsockname()[0]
    return r


def hostname():
    return socket.gethostname()


def str_bool(raw):
    if raw and raw.lower() in ('true', 't', '1'):
        return True
    return False



class Envars:
    def __init__(self):
        load_dotenv()
        self.SCHEMA = os.getenv('APP_SCHEMA', 'public')
        self.APP_ENV = os.getenv('APP_ENV', 'prod')
        self.APP_DEBUG = str_bool(os.getenv('APP_DEBUG'))
        self.APP_PORT = int(os.getenv('APP_PORT', 5000))
        self.CLIENT_ID = os.getenv('CLIENT_ID', hostname())
        self.CLIENT_IP = os.getenv('CLIENT_IP', local_ip())
        self.APP_NAME = os.getenv('APP_NAME', 'app')
        self.IS_WORKER = str_bool(os.getenv('IS_WORKER'))
        self.CONFIG_PATH = '/'.join([os.getenv('CONSUL_KV_PATH', 'config'), self.APP_NAME.lower(), self.APP_ENV.lower()])
        self.CACHE_PATH = Path(os.getenv('CACHE_PATH', '/tmp')).resolve()
        self.LOG_LEVEL = os.getenv('LOG_LEVEL', 'INFO')
        logging.basicConfig(level=self.LOG_LEVEL)


# 兼容旧版本
global_envars = Envars()
# 兼容旧版本的全局变量
SCHEMA = global_envars.SCHEMA
APP_ENV = global_envars.APP_ENV
APP_DEBUG = global_envars.APP_DEBUG
APP_PORT = global_envars.APP_PORT
CLIENT_ID = global_envars.CLIENT_ID
CLIENT_IP = global_envars.CLIENT_IP
APP_NAME = global_envars.APP_NAME
IS_WORKER = global_envars.IS_WORKER
CONFIG_PATH = global_envars.CONFIG_PATH
CACHE_PATH = global_envars.CACHE_PATH
LOG_LEVEL = global_envars.LOG_LEVEL





__all__ = (
  'Envars',
  'load_dotenv',
  'global_envars',
  'SCHEMA',
  'APP_ENV',
  'APP_DEBUG',
  'APP_PORT',
  'CLIENT_ID',
  'CLIENT_IP',
  'APP_NAME',
  'IS_WORKER',
  'CONFIG_PATH',
  'CACHE_PATH',
  'LOG_LEVEL',
)



# Redis driver
# please install redis-py
#
import asyncio
import time
from datetime import timedelta
from functools import wraps
from typing import Any, Optional, Protocol, Union
from uuid import uuid4

from redis.asyncio import Redis
from redis.asyncio.cluster import RedisCluster

__clients = {}


class RedisClientP(Protocol):
    async def exists(self, key) -> bool: ...
    async def get(self, key) -> Any: ...
    async def set(self, key, value, ex: Optional[int] = None) -> bool: ...
    async def setnx(self, key, value) -> int: ...
    async def delete(self, key) -> int: ...
    async def expire(self, key, seconds) -> int: ...


async def new_client(uri: str, tag: str = None):
    if tag is None:
        tag = uuid4()

    if uri.startswith('cluster'):
        __clients[tag] = await RedisCluster.from_url(uri)
    else:
        __clients[tag] = Redis.from_url(uri)


def connect(tag: Union[str, int] = 0) -> RedisClientP:
    pool = __clients.get(tag, None)
    if pool:
        return pool
    raise RuntimeWarning(f'Not found redis client with name: {tag}')


async def load(conf: dict):
    if conf and len(conf) > 0:
        for key in conf:
            await new_client(conf[key], key)
            # print(f'redis: {key} connected')


conn = get_client = connect


class Lock:
    def __init__(self, tag, lock_flg):
        self._client = connect(tag)
        self._flg = lock_flg
        self._owner = False

    async def acquire(self, blocking=True, timeout=None):
        if not blocking and timeout is not None:
            raise ValueError("can't specify timeout for non-blocking acquire")
        start_time = int(time.time())
        while await self._client.exists(self._flg):
            if not blocking:
                break
            await asyncio.sleep(0.5)
            curr_time = int(time.time())
            if isinstance(timeout, int) and curr_time - start_time > timeout:
                return False
        r = await self._client.setnx(self._flg, int(time.time()))
        if r:
            self._owner = True
            return True
        return False

    async def release(self):
        if self._owner:
            await self._client.delete(self._flg)
            self._owner = False

    @property
    def locked(self):
        return self._owner


class SpinLock(Lock):
    def __init__(self, tag, flag, blocking=True, timeout=None):
        super(SpinLock, self).__init__(tag, flag)
        self._blocking = blocking
        self._timeout = timeout

    async def __aenter__(self):
        r = await self.acquire(self._blocking, self._timeout)
        if not r:
            raise RuntimeError(f'Cannot acquire lock: {self._flg}')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.release()


class IntervalLimitError(RuntimeError):
    pass


def interval_lock(flg, name, timeout=timedelta(seconds=3)):
    def limit_ann(fn):
        @wraps(fn)
        async def wrapped(*args, **kwargs):
            client = connect(flg)
            time_limit = int(timeout.total_seconds())
            if await client.exists(name):
                raise IntervalLimitError(
                    f'time interval limit in {time_limit}s')
            await client.setnx(name, time.time())
            await client.expire(name, time_limit)
            return await fn(*args, **kwargs)
        return wrapped
    return limit_ann



import os
import platform
from datetime import timedelta
from sys import version_info

if version_info.major >= 3 and version_info.minor > 7:
    from typing import Protocol, runtime_checkable

    @runtime_checkable
    class Writable(Protocol):
        def write(self, *args, **kwargs): ...
        def close(self): ...


    @runtime_checkable
    class Readable(Protocol):
        def read(self, *args, **kwargs): ...
        def close(self): ...
else:
    class Writable(object):
        def write(self, *args, **kwargs): ...
        def close(self): ...


    class Readable(object):
        def read(self, *args, **kwargs): ...
        def close(self): ...


CACHE_DEFAULT_DIR = os.getenv('TEMP') if platform.system() == 'Windows' else '/tmp'
DEFAULT_CACHE_TIME = timedelta(seconds=30)  # 30 seconds



import asyncio
import logging
import os

import pytest
import pytest_asyncio
import yaml

from jcutil.drivers.redis import Lock, SpinLock, connect, load, new_client


@pytest_asyncio.fixture
async def setup_redis():
    """设置Redis测试连接"""
    # 如果存在测试配置文件，则使用配置文件
    if os.path.exists('tests/config.yaml'):
        with open('tests/config.yaml', 'r') as f:
            conf = yaml.safe_load(f)
            if 'redis' in conf:
                await load(conf['redis'])
                return

    # 否则使用默认连接
    await new_client('redis://127.0.0.1:6379', 'test')


@pytest.mark.asyncio
async def test_redis_basic_operations(setup_redis):
    """测试Redis基本操作"""
    # 获取连接
    client = connect('test')

    # 测试设置和获取值
    test_key = 'test_key'
    test_value = 'test_value'

    # 确保键不存在
    await client.delete(test_key)

    # 测试设置值
    assert await client.set(test_key, test_value) is True

    # 测试获取值
    result = await client.get(test_key)
    assert result.decode('utf-8') == test_value

    # 测试键存在
    assert await client.exists(test_key) == 1

    # 测试删除键
    assert await client.delete(test_key) == 1

    # 确认键已删除
    assert await client.exists(test_key) == 0


@pytest.mark.asyncio
async def test_redis_expire(setup_redis):
    """测试Redis过期时间设置"""
    client = connect('test')
    test_key = 'expire_key'

    # 设置带过期时间的键
    await client.set(test_key, 'temporary', ex=2)

    # 键应该存在
    assert await client.exists(test_key) == 1

    # 等待过期
    await asyncio.sleep(3)

    # 键应该已过期
    assert await client.exists(test_key) == 0


@pytest.mark.asyncio
async def test_spin_lock(setup_redis):
    """测试SpinLock功能"""
    lock_key = 'test_lock'

    # 测试自动获取和释放锁
    async with SpinLock('test', lock_key):
        # 在锁内执行操作
        await asyncio.sleep(1)
        logging.info('holding lock for 1 second')

    # 测试手动锁操作
    lock = Lock('test', lock_key)
    acquired = await lock.acquire()
    assert acquired is True
    assert lock.locked is True

    # 尝试获取已被锁定的锁
    lock2 = Lock('test', lock_key)
    acquired2 = await lock2.acquire(blocking=False)
    assert acquired2 is False

    # 释放锁
    await lock.release()
    assert lock.locked is False

    # 现在应该能获取锁了
    acquired2 = await lock2.acquire(blocking=False)
    assert acquired2 is True

    # 释放第二个锁
    await lock2.release()




import pandas as pd

from jcutil.core import async_run
from jcutil.drivers import db


async def read_sql(schema, sql, **kwargs):
  with db.get_client(schema) as engine:
    return await async_run(pd.read_sql, sql, engine, **kwargs)


async def read_table(schema, table_name, **kwargs):
  with db.get_client(schema) as engine:
    return await async_run(pd.read_sql_table, table_name, engine, **kwargs)





from html import escape

from jcramda import when
from jcramda.base.comparison import is_a_str, is_a_tuple


def _escape(raw):
  if raw.startswith('{{') and raw.endswith('}}'):
    return raw.strip('{{').strip('}}')
  if is_a_str(raw):
    return f"'{escape(raw)}'"
  return str(raw)


_value_check = when(
  (is_a_str, lambda s: f' = {_escape(s)}'),
  (is_a_tuple, lambda tp: f'{tp[0]} {_escape(tp[1])}'),
  else_=str
)


class Where:
  def __init__(self, raw):
      self.raw = raw

  def __str__(self):
    tn = getattr(type(self.raw), '__name__')
    method = getattr(self, f'from_{tn}')
    if callable(method):
      return method()
    else:
      return str(self.raw)

  def from_dict(self):
    """
    Example:
    >>> w = where({'name': 'someone', 'age': ('>', 32), 'birthday': '{{CURRENT_DATE}}'})
    >> print(w)
       name = 'someone' AND 'age' > 32 AND birthday = CURRENT_DATE
    """
    return ' AND '.join([f'{key} {_value_check(value)}' for key, value in self.raw.items()])




from enum import IntEnum
from typing import AnyStr, Tuple

BS: int

Model: IntEnum


def aes_encrypt(key: AnyStr, plain: str, /, mode: Model = ...) -> Tuple[str, str]: ...


def aes_decrypt(
  key: AnyStr, value: str, /,
  mode: Model = ..., nonce_or_iv: str = ...) -> str: ...


def get_sha1prng_key(key: str) -> str: ...


def aes_ecb_encrypt(key: AnyStr, plaintext: str) -> str: ...
def aes_ecb_decrypt(key: AnyStr, ciphertext: str) -> str: ...
def aes_cfb_encrypt(key: AnyStr, plaintext: str) -> Tuple[str, str]: ...
def aes_cfb_decrypt(key: AnyStr, ciphertext: str, /, iv: str) -> str: ...

def to_base64(hexstring: str) -> str: ...
def sha3sum(plaintext: str) -> str: ...



import time

import pandas as pd

from jcutil.drivers import db


def quick_update(schema, target: str, set_cols, where_cols, data, /,
                 mode='update'):
    """利用pandas快速生成的临时表来批量更新数据

    **使用RPLACE模式需要注意，如果传入的数据字段少于原有字段，则原表中的这些字段将被至为NULL**

    Args:
        schema (str): Database bind alias name
        target (str): update target table name
        set_cols (str, list[str]): Will be updated columns. 'id,name,somecol' or ['id','name','somecol']
        where_cols (str, list[str]): Columns that be used in where. 'id,name,somecol' or ['id', 'name', 'somecol']
        data (Iterable, DataFrame): a data that can pass to pd.DataFrame
        mode (str, optional): sql mode: ['update', 'replace']. Defaults to 'update'.

    Exceptions:
        ValueError: some column in set_cols or where_cols which is not in target table.
    """
    temp_name = f'tmp_{int(time.time())}'
    if isinstance(where_cols, (str, bytes)):
        where_cols = where_cols.split(',')
    if isinstance(set_cols, (str, bytes)):
        set_cols = set_cols.split(',')
    target_columns = set(pd.read_sql(f'describe {target}', db.conn(schema)).Field.values)
    df = pd.DataFrame(data)
    if diff := (set(set_cols) - target_columns):
        raise ValueError(f'set_cols: {diff} columns not in target table.')
    if diff := (set(where_cols) - target_columns):
        raise ValueError(f'where_cols: {diff} columns not in target table.')
    cols = [col for col in df.columns if col in target_columns]

    if df.empty or not cols:
        return
    if mode == 'update':
        sql = 'UPDATE {target} t, {source} s SET {sets} WHERE {where}'.format(
            target=target,
            source=temp_name,
            sets=','.join([f't.{c} = s.{c}' for c in set_cols]),
            where=' AND '.join([f't.{c} = s.{c}' for c in where_cols]),
        )
    else:
        sql = 'REPLACE INTO {target} ({sets}) SELECT {sets} FROM {source}'.format(
            target=target, source=temp_name,
            sets=','.join(cols)
        )
    with db.conn(schema).begin() as session:
        session.execute(f'create table {temp_name} like {target}')
        df[cols].to_sql(temp_name, session, index=False, if_exists='append')
        session.execute(sql)
        session.execute(f'drop table {temp_name}')
        session.commit()



import os
import sys
from enum import IntEnum
from typing import Any, List, Optional, Tuple, Union

from colorama import init
from jcramda import first, join, partial, zip_

from jcutil.core import c_write, nl_print

__all__ = (
    'Color', 'FontFormat', 'Chalk', 'EndFlag', 'RedChalk', 'GreenChalk', 'BlackChalk',
    'BlueChalk', 'MagentaChalk', 'WhiteChalk', 'YellowChalk', 'CyanChalk',
    'show_menu', 'select', 'BoldChalk',
    'BrightRedChalk', 'BrightBlueChalk', 'BrightCyanChalk', 'BrightBlackChalk', 'BrightGreenChalk',
    'BrightWhiteChalk', 'BrightYellowChalk', 'BrightMagentaChalk',
)


__CHALK_TMPL__ = '\033[{}m'
__RESET__ = '\033[0m'

# 在Windows系统上初始化colorama
if os.name == 'nt':
    init(convert=True, strip=False)


class Color(IntEnum):
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 39
    BRIGHT_BLACK = 90
    BRIGHT_RED = 91
    BRIGHT_GREEN = 92
    BRIGHT_YELLOW = 93
    BRIGHT_BLUE = 94
    BRIGHT_MAGENTA = 95
    BRIGHT_CYAN = 96
    BRIGHT_WHITE = 97

    @property
    def bgcolor(self) -> int:
        """获取对应的背景色值"""
        return self.value + 10

    @property
    def echo(self):
        """获取带有当前颜色的Chalk函数"""
        return partial(Chalk, fgc=self)


class FontFormat(IntEnum):
    BOLD = 1
    LIGHT = 2
    ITALIC = 3
    UNDER_LINE = 4
    BLINK = 5
    RESERVE = 7
    DELETE = 9


class EndFlag(IntEnum):
    ALL_END = 0
    B_END = 22
    I_END = 23
    UL_END = 24
    BL_END = 25
    R_END = 27
    D_END = 29


def _gen_raw(fgc: Optional[Color] = None, bgc: Optional[Color] = None, *styles: FontFormat) -> str:
    """生成ANSI转义序列

    Args:
        fgc: 前景色
        bgc: 背景色
        styles: 字体样式

    Returns:
        ANSI转义序列字符串
    """
    # 过滤None值，确保join函数收到的是有效值
    raw = [value for value in (
        fgc.value if fgc else None,
        bgc.bgcolor if bgc else None,
        *[ff.value for ff in styles if ff is not None]
    ) if value is not None]

    if not raw:
        return __RESET__

    return __CHALK_TMPL__.format(join(';')(raw))


class Chalk:
    """控制台彩色文本工具类"""

    def __init__(self, text: Optional[str] = None,
                 fgc: Optional[Color] = None,
                 bgc: Optional[Color] = None,
                 styles: Tuple[FontFormat, ...] = ()):
        """初始化Chalk实例

        Args:
            text: 要设置样式的文本
            fgc: 前景色
            bgc: 背景色
            styles: 字体样式
        """
        self.__chains__: List[str] = [_gen_raw(fgc, bgc, *styles)]
        self.__buffer__: List[str] = []
        if text:
            self.text(text)

    def use(self, *args: FontFormat, **kwargs) -> 'Chalk':
        """使用指定样式

        Args:
            *args: 字体样式
            **kwargs: 关键字参数
                fg_color: 前景色
                bg_color: 背景色

        Returns:
            Chalk实例
        """
        self.__chains__.append(_gen_raw(
            kwargs.get('fg_color'), kwargs.get('bg_color'), *args))
        return self

    def end(self, *flag: EndFlag) -> 'Chalk':
        """结束当前样式

        Args:
            *flag: 结束标志

        Returns:
            Chalk实例
        """
        if len(flag) > 0:
            self.__chains__.append(_gen_raw(None, None, *flag))
        else:
            self.__chains__.append(__RESET__)
        return self

    def _text(self, text: str) -> str:
        """处理文本，替换重置符号

        Args:
            text: 待处理文本

        Returns:
            处理后的文本
        """
        if not text:
            return ""
        return str(text).replace(__RESET__, first(self.__chains__) or '')

    def text(self, text: str) -> 'Chalk':
        """添加文本

        Args:
            text: 要添加的文本

        Returns:
            Chalk实例
        """
        self.__buffer__.append(self._text(text))
        return self

    def format(self, text: str, *style: FontFormat) -> None:
        """格式化文本

        Args:
            text: 文本内容
            *style: 样式
        """
        self.use(*style).text(text).end(EndFlag.ALL_END)

    def bold(self, text: str) -> 'Chalk':
        """添加粗体文本

        Args:
            text: 文本内容

        Returns:
            Chalk实例
        """
        self.use(FontFormat.BOLD).text(text)
        return self

    def italic(self, text: str) -> 'Chalk':
        """添加斜体文本

        Args:
            text: 文本内容

        Returns:
            Chalk实例
        """
        self.use(FontFormat.ITALIC).text(text)
        return self

    def underline(self, text: str) -> 'Chalk':
        """添加下划线文本

        Args:
            text: 文本内容

        Returns:
            Chalk实例
        """
        self.use(FontFormat.UNDER_LINE).text(text)
        return self

    def expandtabs(self) -> str:
        """展开制表符

        Returns:
            展开制表符后的文本
        """
        return str(self).expandtabs()

    @property
    def raw(self) -> str:
        """获取原始文本（不含样式）

        Returns:
            原始文本
        """
        return ''.join(self.__buffer__)

    def __str__(self) -> str:
        """获取完整的彩色文本

        Returns:
            带有ANSI控制序列的文本
        """
        # 使用zip_来配对chains和buffer
        data = zip_('', self.__chains__, self.__buffer__)
        return ''.join([f'{c}{t}' for c, t in data]) + __RESET__

    def __len__(self) -> int:
        """获取文本长度

        Returns:
            文本长度（不含样式）
        """
        return len(self.raw)

    def __add__(self, other: Union[str, 'Chalk']) -> Union[str, 'Chalk']:
        """连接两个Chalk对象或Chalk与字符串

        Args:
            other: 另一个Chalk对象或字符串

        Returns:
            连接后的对象
        """
        if isinstance(other, str):
            return str(self) + other

        new_chalk = Chalk()
        new_chalk.__chains__ = self.__chains__ + [__RESET__] + other.__chains__
        new_chalk.__buffer__ = self.__buffer__ + [''] + other.__buffer__
        return new_chalk

    def __radd__(self, other: str) -> str:
        """实现右侧加法

        Args:
            other: 字符串

        Returns:
            连接后的字符串
        """
        if isinstance(other, str):
            return other + str(self)
        return NotImplemented

    def __repr__(self) -> str:
        """获取对象的字符串表示

        Returns:
            对象的调试表示
        """
        return f'Chalk<buff:{self.__buffer__}, chain:{self.__chains__}>'

    def __call__(self, *args, **kwargs) -> str:
        """调用对象时返回字符串表示

        Returns:
            字符串表示
        """
        return str(self)

    def __mod__(self, args) -> str:
        """支持字符串格式化

        Args:
            args: 格式化参数

        Returns:
            格式化后的字符串
        """
        return str(self) % args


# 预定义的Chalk实例
RedChalk = partial(Chalk, fgc=Color.RED)
GreenChalk = partial(Chalk, fgc=Color.GREEN)
BlueChalk = partial(Chalk, fgc=Color.BLUE)
YellowChalk = partial(Chalk, fgc=Color.YELLOW)
MagentaChalk = partial(Chalk, fgc=Color.MAGENTA)
CyanChalk = partial(Chalk, fgc=Color.CYAN)
WhiteChalk = partial(Chalk, fgc=Color.WHITE)
BlackChalk = partial(Chalk, fgc=Color.BLACK)
BoldChalk = partial(Chalk, styles=(FontFormat.BOLD,))

BrightBlackChalk = partial(Chalk, fgc=Color.BRIGHT_BLACK)
BrightBlueChalk = partial(Chalk, fgc=Color.BRIGHT_BLUE)
BrightCyanChalk = partial(Chalk, fgc=Color.BRIGHT_CYAN)
BrightGreenChalk = partial(Chalk, fgc=Color.BRIGHT_GREEN)
BrightMagentaChalk = partial(Chalk, fgc=Color.BRIGHT_MAGENTA)
BrightRedChalk = partial(Chalk, fgc=Color.BRIGHT_RED)
BrightYellowChalk = partial(Chalk, fgc=Color.BRIGHT_YELLOW)
BrightWhiteChalk = partial(Chalk, fgc=Color.BRIGHT_WHITE)


def show_menu(items: List[tuple], is_submenu: bool = False, title: Union[str, Chalk] = '命令菜单') -> Any:
    """显示菜单并获取用户选择

    Args:
        items: 菜单项列表，每项为(标题, 函数)的元组
        is_submenu: 是否为子菜单
        title: 菜单标题

    Returns:
        所选菜单项的函数结果
    """
    if not items:
        return None

    items_copy = items.copy()
    index = 1

    # 显示标题
    if isinstance(title, Chalk):
        c_write(title, end='\n\n')
    else:
        c_write(GreenChalk(title, styles=(FontFormat.BOLD,)), end='\n\n')

    # 添加返回和退出选项
    if is_submenu:
        items_copy.append(('返回', lambda: ''))
    items_copy.append(('退出', sys.exit))

    # 显示菜单项
    for item_title, _ in items_copy:
        sys.stdout.write(
            YellowChalk(f'{index}. {item_title}', styles=(FontFormat.BOLD,))() + '\n')
        index += 1

    nl_print('\n')

    # 获取用户选择并执行
    try:
        choice = select()
        if 1 <= choice <= len(items_copy):
            return items_copy[choice - 1][1]()
        else:
            print(RedChalk("无效选择，请重试。").bold("!"))
            return show_menu(items, is_submenu, title)
    except ValueError:
        print(RedChalk("请输入有效数字").bold("!"))
        return show_menu(items, is_submenu, title)


def select() -> int:
    """等待用户输入并返回所选数字

    Returns:
        用户选择的数字
    """
    c_write(GreenChalk('选择功能：')())
    sys.stdout.flush()
    try:
        # 支持更大的菜单项数量
        user_input = sys.stdin.readline().strip()
        return int(user_input)
    except ValueError:
        return 0  # 返回无效值，让调用者处理



import pytest

from jcutil.drivers.mongo import MongoClient, get_client, new_client

# 测试MongoDB连接
# 运行测试前请确保有MongoDB服务器运行在本地27017端口
TEST_MONGO_URI = "mongodb://localhost:27017/test_db"
TEST_COLLECTION = "test_collection"


@pytest.fixture(scope="module")
def client():
    client = MongoClient(TEST_MONGO_URI, "test_client")
    yield client
    # 清理测试数据
    db = client.get_database()
    db[TEST_COLLECTION].drop()
    # 关闭连接以防内存泄漏
    client.sync_client.close()
    # Don't close async_client here as it will close the event loop
    # client.async_client.close()


def test_sync_api(client):
    # 测试同步API
    # 插入测试数据
    data = {"name": "test_item", "value": 100}
    result = client.save(TEST_COLLECTION, data)
    assert "_id" in result
    assert result["name"] == "test_item"

    # 查询数据
    found = client.find_by_id(TEST_COLLECTION, result["_id"])
    assert found["name"] == "test_item"

    # 更新数据
    data["value"] = 200
    updated = client.save(TEST_COLLECTION, data)
    assert updated["value"] == 200

    # 查询所有数据
    all_items = client.find(TEST_COLLECTION, {})
    assert len(all_items) > 0

    # 删除数据
    delete_result = client.delete(TEST_COLLECTION, result["_id"])
    assert delete_result == 1


@pytest.mark.asyncio
async def test_async_api(client):
    # 测试异步API
    # 插入测试数据
    data = {"name": "async_test_item", "value": 300}
    result = await client.async_save(TEST_COLLECTION, data)
    assert "_id" in result
    assert result["name"] == "async_test_item"

    # 查询数据
    found = await client.async_find_by_id(TEST_COLLECTION, result["_id"])
    assert found["name"] == "async_test_item"

    # 更新数据
    data["value"] = 400
    updated = await client.async_save(TEST_COLLECTION, data)
    assert updated["value"] == 400

    # 查询所有数据
    all_items = await client.async_find(TEST_COLLECTION, {})
    assert len(all_items) > 0

    # 删除数据
    delete_result = await client.async_delete(TEST_COLLECTION, result["_id"])
    assert delete_result == 1


@pytest.mark.asyncio
async def test_proxy(client):
    # 测试同步代理
    proxy = client.create_proxy(TEST_COLLECTION)
    data = {"name": "proxy_test", "value": 500}
    added = proxy.add(data)
    assert added["name"] == "proxy_test"

    # 存储添加的数据ID，用于后续清理
    added_id = added["_id"]

    try:
        # 测试异步代理
        async_proxy = await client.create_async_proxy(TEST_COLLECTION)
        async_data = {"name": "async_proxy_test", "value": 600}
        async_added = await async_proxy.add(async_data)
        assert async_added["name"] == "async_proxy_test"

        # 存储异步添加的数据ID，用于后续清理
        async_added_id = async_added["_id"]

        # 查询并验证
        items = proxy.all()
        assert len(items) >= 2

        async_items = await async_proxy.all()
        assert len(async_items) >= 2

        # 清理异步测试数据
        await async_proxy.delete(async_added_id)
    finally:
        # 无论测试成功与否，都要清理同步测试数据
        proxy.delete(added_id)


def test_global_client_management():
    # 测试全局客户端管理
    new_client(TEST_MONGO_URI, "global_test")
    client = get_client("global_test")

    # 测试基本操作
    data = {"name": "global_test_item", "value": 700}
    result = client.save(TEST_COLLECTION, data)
    assert result["name"] == "global_test_item"

    # 清理
    client.delete(TEST_COLLECTION, result["_id"])



import os

import pytest
from jcramda import hex_token, members

from jcutil.crypto import Mode, aes_decrypt, aes_encrypt, sha3sum, to_base64


@pytest.fixture()
def modes():
    return members(Mode)


def test_aes(modes):
    raw = os.getenv('TXT', 'This is a Test text')
    print('plaintext =', raw, '| hashsum:', sha3sum(raw))
    # genrate a 32 bytes key
    key = hex_token(32)
    print('key:', key)
    for m in modes:
      print(m)
      ciphertext, iv = aes_encrypt(key, raw, mode=m)
      print(ciphertext, iv)
      r = aes_decrypt(key, mode=m, nonce_or_iv=iv)(ciphertext)
      assert r == raw
      print(to_base64(ciphertext))



import pytest

from jcutil.consul import ConfigFormat, KvProperty


class TestA:
  name = KvProperty('name')
  bar = KvProperty('foo', format=ConfigFormat.Yaml, cached=True)

  def desc(self):
    print('my name is:', self.name)


@pytest.mark.skip(reason="需要配置Consul服务才能运行此测试")
def test_kvp():
  ta = TestA()
  ta.desc()
  assert ta.name == 'FooBar'
  print(ta.bar, ta.foo)



import logging

from jcutil import chalk
from jcutil import consul as kv
from jcutil.drivers import smart_load
from jcutil.server.envars import global_envars

context = {}


def load_config(*args, config_path=global_envars.CONFIG_PATH, v=True):
    """
    初始化配置，并自动加载配置中的各种底层驱动

    **如只需要读取配置信息，请勿使用此方法**
    Parameters
    ----------
    v
    args

    Returns
    -------

    """
    conf = kv.fetch_key(config_path, fmt=kv.ConfigFormat.Yaml)
    if len(args) > 0:
        needed_conf = {}
        for key in ['server', *args]:
            assert key in conf, f'not found [{key}] in kv server.'
            needed_conf[key] = conf[key]
        conf = needed_conf
    v and print(config_path, ':', chalk.GreenChalk(conf))
    try:
        smart_load(conf)
        context['conf'] = conf
    except Exception as err:
        logging.error('read config failed: %s', err)



import datetime as dt
from collections import namedtuple
from decimal import Decimal
from json import JSONDecoder, JSONEncoder, dump, dumps, load, loads
from typing import Any, Iterable
from uuid import UUID

from jcramda import (
    attr,
    b64_encode,
    camelcase,
    compose,
    flat_concat,
    has_attr,
    identity,
    is_a,
    is_a_int,
    is_a_mapper,
    key_map,
    partial,
    when,
)

from .pdtools import TYPE_REGS

_str_to_type = {
    'true': True,
    'false': False,
}

__all__ = (
    'SafeJsonEncoder',
    'SafeJsonDecoder',
    'to_json',
    'to_json_file',
    'pp_json',
    'fix_document',
    'to_obj',
    'from_json_file',
)


_type_regs = (
    *TYPE_REGS,
    (is_a((UUID, )), str),
    (is_a(dt.datetime), lambda o: o.strftime('%Y-%m-%d %H:%M:%S')),
    (is_a(bytes), b64_encode),
    (is_a(memoryview), compose(b64_encode, bytes)),
    (is_a(dict), flat_concat),
    (is_a_int, int),
    (has_attr('__html__'), compose(identity, attr('__html__'))),
    (is_a(str), lambda s: _str_to_type.get(s, s))
)


class SafeJsonEncoder(JSONEncoder):

    def default(self, object_: Any) -> Any:
        r = when(*_type_regs, str)(object_)
        return key_map(camelcase, r)


class SafeJsonDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=fix_document, *args, **kwargs)


to_json = partial(dumps, cls=SafeJsonEncoder, ensure_ascii=False)

def to_json_file(obj, fp, **kwargs):
    """将对象序列化为JSON并写入文件

    Args:
        obj: 要序列化的对象
        fp: 文件路径或文件对象
        **kwargs: 其他传递给json.dump的参数
    """
    kwargs.setdefault('cls', SafeJsonEncoder)
    kwargs.setdefault('ensure_ascii', False)

    if isinstance(fp, str):
        with open(fp, 'w', encoding='utf-8') as f:
            dump(obj, f, **kwargs)
    else:
        dump(obj, fp, **kwargs)


DocFixedOpt = namedtuple('DocFixedOpt', 'where, fixed')


def fix_document(doc, fix_options: Iterable[DocFixedOpt] = ()):
    if is_a_mapper(doc):
        r = {}
        for k, v in doc.items():
            new_key, new_v = k, v
            if str(k).startswith('$'):
                if len(doc) == 1:
                    return fix_document(v, fix_options)
                new_key = k[1:]
            r[new_key] = fix_document(new_v, fix_options)
        return r
    elif is_a((list, tuple, set), doc):
        return [fix_document(x, fix_options) for x in doc]

    if str(doc).lower() in ('nan', 'nat', 'null'):
        return None

    return when(
        *fix_options,
        (is_a(Decimal), identity),
        else_=doc
    )(doc)


to_obj = partial(loads, cls=SafeJsonDecoder)


def from_json_file(file_path):
    with open(file_path, 'r') as fp:
        s = load(fp, cls=SafeJsonDecoder)
    if is_a(str, s):
        s = to_obj(s)
    return s


def pp_json(obj):
    printed_str = dumps(obj, indent=2, ensure_ascii=False)
    try:
        from pygments import formatters, highlight, lexers
        colorful_json = highlight(printed_str, lexers.JsonLexer(),
                                formatters.TerminalFormatter())
    except ModuleNotFoundError:
        from jcutil.chalk import GreenChalk
        colorful_json = GreenChalk(printed_str)
    return colorful_json



"""
测试Chalk模块 - 用于控制台彩色输出

这个测试文件验证jcutil.chalk模块的功能是否正常工作，包括颜色输出、样式设置、文本操作等。
"""
import io
import sys

import pytest

from jcutil.chalk import (
    BlueChalk,
    Chalk,
    Color,
    EndFlag,
    FontFormat,
    GreenChalk,
    RedChalk,
    YellowChalk,
)


@pytest.fixture
def capture_stdout():
    """捕获标准输出的fixture"""
    captured_output = io.StringIO()
    old_stdout = sys.stdout
    sys.stdout = captured_output

    yield captured_output

    # 恢复标准输出
    sys.stdout = old_stdout
    # 打印捕获的输出，方便调试
    print(captured_output.getvalue())


def test_chalk_init(capture_stdout):
    """测试Chalk初始化和链式调用"""
    red = Chalk('hello', Color.RED)
    assert len(red.__chains__) > 0
    print(red)
    assert len(red) > 0

    green = GreenChalk('oh, it is a ').use(FontFormat.BOLD).text('green').end(EndFlag.B_END).text(' chalk.')
    print(repr(green))
    print(green)

    merge = red + green
    print(repr(merge))
    print(merge)


def test_add(capture_stdout):
    """测试加法操作符重载"""
    red = RedChalk('hello')
    r = red + ' world'
    assert isinstance(r, str), 'return a str when add a str'
    assert r == '\033[31mhello\033[0m world'
    print(r)

    r = red + GreenChalk('|Mo')
    assert str(r) == '\033[31mhello\033[0m\033[32m|Mo\033[0m'
    print(r)


def test_mod(capture_stdout):
    """测试格式化操作符重载"""
    red = RedChalk('hello %s')
    print(red)
    r = red % 'world'
    assert r == '\033[31mhello world\033[0m'
    print(r)
    print(red % 111)


def test_wrapper(capture_stdout):
    """测试嵌套使用功能"""
    red = RedChalk('[wappered]')
    r = GreenChalk(f'a {red} b')
    print(repr(r))
    print(r)

    br = YellowChalk().bold('bold string')
    print(repr(br), br)


def test_new_features(capture_stdout):
    """测试新增的功能"""
    # 测试italic方法
    italic_text = RedChalk().italic("This is italic text")
    print(italic_text)
    assert '\033[3m' in str(italic_text), "斜体标记应存在"

    # 测试underline方法
    underline_text = BlueChalk().underline("This is underlined text")
    print(underline_text)
    assert '\033[4m' in str(underline_text), "下划线标记应存在"

    # 测试__radd__方法
    text_with_chalk = "Plain text with " + GreenChalk("green text")
    print(text_with_chalk)
    assert text_with_chalk.endswith('\033[0m'), "应以重置颜色标记结尾"
    assert text_with_chalk.startswith('Plain text with '), "应以普通文本开始"

    # 测试raw属性
    colored_text = RedChalk("Hello").bold(" World")
    print(colored_text)
    assert colored_text.raw == "Hello World"


if __name__ == "__main__":
    print(YellowChalk().bold("=== 开始测试Chalk模块 ===").expandtabs())
    pytest.main(["-xvs", __file__])



"""
测试jcutil core模块的功能

此测试文件验证jcutil.core模块中的核心功能，包括JSON处理、对象序列化和异步操作等。
"""
import tempfile
from pathlib import Path

import pytest

from jcutil.core import (
    async_run,
    from_json_file,
    map_async,
    obj_dumps,
    obj_loads,
    to_json,
    to_json_file,
    to_obj,
)


def test_json_serialization():
    """测试JSON序列化功能"""
    # 测试基本数据类型
    data = {
        "string": "hello",
        "number": 42,
        "boolean": True,
        "null": None,
        "array": [1, 2, 3],
        "object": {"a": 1, "b": 2}
    }

    # 测试to_json
    json_str = to_json(data)
    assert isinstance(json_str, str)

    # 测试反序列化
    obj = to_obj(json_str)
    assert obj["string"] == "hello"
    assert obj["number"] == 42

    # 测试特殊处理器
    complex_data = {
        "date": "2023-01-01",
        "camelCaseKey": "value",
    }
    json_str = to_json(complex_data)
    assert "camelCaseKey" in json_str


def test_json_file_operations():
    """测试JSON文件操作功能"""
    data = {"test": "file_operation", "value": 123}

    # 创建临时文件
    with tempfile.NamedTemporaryFile(suffix='.json', delete=False) as temp_file:
        temp_path = temp_file.name

    try:
        # 测试to_json_file
        to_json_file(data, temp_path)

        # 测试from_json_file
        loaded_data = from_json_file(temp_path)
        assert loaded_data["test"] == "file_operation"
        assert loaded_data["value"] == 123
    finally:
        # 清理临时文件
        Path(temp_path).unlink(missing_ok=True)


@pytest.mark.asyncio
async def test_async_functions():
    """测试异步功能"""
    def sync_func(x):
        return x * 2

    # 测试async_run
    result = await async_run(sync_func, 21)
    assert result == 42

    # 测试带上下文的async_run
    result_with_context = await async_run(sync_func, 21, with_context=True)
    assert result_with_context == 42


def test_map_async():
    """测试并行映射功能"""
    data = [1, 2, 3, 4, 5]

    def multiply_by_2(x):
        return x * 2

    # 测试map_async
    results = map_async(multiply_by_2, data)
    assert sorted(results) == [2, 4, 6, 8, 10]


def test_obj_serialization():
    """测试对象序列化功能"""
    # 创建一个复杂对象
    obj = {"complex": [1, 2, 3], "nested": {"a": 1, "b": 2}}

    # 测试序列化和反序列化
    serialized = obj_dumps(obj)
    assert isinstance(serialized, bytes)

    deserialized = obj_loads(serialized)
    assert deserialized["complex"] == [1, 2, 3]
    assert deserialized["nested"]["a"] == 1



import asyncio
from enum import Enum
from typing import Optional

from apscheduler.jobstores.mongodb import MongoDBJobStore
from apscheduler.schedulers.base import BaseJobStore, BaseScheduler
from jcramda import depop, from_import_as

from .core import host_mac

__all__ = (
    'SchedulerType',
    'create_by_mongo',
    'scheduler',
    'job2dict',
    'default_store_opts',
)


ERR_MSG = 'scheduler is not created'


def job2dict(job):
    if not job:
        return {}
    return {
        'id': job.id,
        'name': job.name,
        'trigger': job.trigger,
        'func': job.func.__name__,
        'executor': job.executor,
        'args': job.args,
        'kwargs': job.kwargs,
        'max_instances': job.max_instances,
        'next_run_time': job.next_run_time,
        'misfire_grace_time': job.misfire_grace_time,
        'coalesce': job.coalesce,
    }


class SchedulerManager(object):
    instance: Optional[BaseScheduler]
    store: Optional[BaseJobStore]
    lock: asyncio.Lock

    def __init__(self, cron: BaseScheduler = None):
        self.lock = asyncio.Lock()
        self.instance = cron

    async def start(self):
        cron = self.instance
        assert cron, ERR_MSG
        assert not cron.running, 'scheduler is already running'
        await scheduler.lock.acquire()
        cron.start()
        while scheduler.lock.locked():
            await asyncio.sleep(1)

    def stop(self):
        assert self.instance, ERR_MSG
        assert self.instance.running, 'scheduler is not running'
        self.instance.shutdown()
        scheduler.lock.release()

    def jobs(self):
        assert self.instance, ERR_MSG
        if not self.instance.running:
            self.instance.start(True)
        r = list(map(job2dict, self.store.get_all_jobs()))
        return r

    def add_job(self, func, job_opts):
        """

        Parameters
        ----------
        func
        job_opts: dict
            "id": job id, Opitonal.
            "name": job name
            "args": func args
            "kwargs": func kwargs
            "trigger": ``date``, ``interval`` or ``cron``

            ## date params
                - datetime|str run_date: the date/time to run the job at
                - datetime.tzinfo|str timezone: time zone for ``run_date`` if it doesn't have one already

            ## interval params
               - int weeks: number of weeks to wait
               - int days: number of days to wait
               - int hours: number of hours to wait
               - int minutes: number of minutes to wait
               - int seconds: number of seconds to wait
               - datetime|str start_date: starting point for the interval calculation
               - datetime|str end_date: latest possible date/time to trigger on
               - datetime.tzinfo|str timezone: time zone to use for the date/time calculations
               - int|None jitter: advance or delay the job execution by ``jitter`` seconds at most.

            ## cron params
                - int|str year: 4-digit year
                - int|str month: month (1-12)
                - int|str day: day of the (1-31)
                - int|str week: ISO week (1-53)
                - int|str day_of_week: number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
                - int|str hour: hour (0-23)
                - int|str minute: minute (0-59)
                - int|str second: second (0-59)
                - datetime|str start_date: earliest possible date/time to trigger on (inclusive)
                - datetime|str end_date: latest possible date/time to trigger on (inclusive)
                - datetime.tzinfo|str timezone: time zone to use for the date/time calculations
                           (defaults to scheduler timezone)
                - int|None jitter: advance or delay the job execution by ``jitter`` seconds at most.

                .. note:: The first weekday is always **monday**.

        Returns
        -------

        """
        assert self.instance, 'scheduler is not created'
        return self.instance.add_job(func, **job_opts)

    def modify(self, job_id, change):
        job = self.store.lookup_job(job_id)
        assert job, f'not found job by id {job_id}'
        if 'args' in change or 'kw' in change:
            args, kw = depop(['args', 'kw'], change)
            job = self.instance.modify_job(job_id, args=args, kwargs=kw)
        if 'trigger' in change:
            job = self.instance.reschedule_job(job_id, **change)
        return job2dict(job)

    def remove(self, job_id):
        try:
            self.instance.remove_job(job_id)
        except RuntimeError:
            return False
        return True

    @property
    def running(self):
        return self.instance.running

    def find_job(self, job_id):
        return job2dict(self.store.lookup_job(job_id))


class SchedulerType(Enum):
    ASYNC = 'apscheduler.schedulers.asyncio:AsyncIOScheduler'
    BACKGROUND = 'apscheduler.schedulers.background:BackgroundScheduler'
    SYNC = 'apscheduler.schedulers.blocking:BlockingScheduler'


scheduler = SchedulerManager()


def create_by_mongo(s_type: SchedulerType = SchedulerType.BACKGROUND,
                    **mongo_store_opts) -> SchedulerManager:
    assert scheduler.instance is None, 'scheduler is already created'
    clazz = from_import_as(s_type.value)
    assert clazz, f'not found scheduler class named: {s_type.value}'
    instance: BaseScheduler = clazz()
    scheduler.store = MongoDBJobStore(**mongo_store_opts)
    instance.add_jobstore(scheduler.store)
    scheduler.instance = instance
    return scheduler


def default_store_opts(node=None) -> dict:
    node_id = node if node else host_mac()
    from .drivers import mongo
    mongo_client = mongo.conn()
    return {
        'collection': f'jobs.{node_id}',
        'database': mongo_client.get_default_database().name,
        'client': mongo_client
    }



from datetime import timedelta
from typing import Any, Callable, NoReturn, TypeVar, Union

from jcutil.defines import Writable

T = TypeVar('T')


def mem_cache(cache_dir: str = ...) -> Callable: ...

def clear_mem(path: str = ..., cache_dir: str = ...) -> NoReturn: ...

def redis_cache(expires: Union[int, timedelta], prefix: str = ...,
                redis_connector: Callable = ...,
                result_assert: Callable[[Any], bool] = ...,
                ) -> Callable[[Callable], Any]: ...

def persistence(fs: Union[Writable, Callable[[], Writable]], f: Callable[..., T]) \
        -> Callable[..., T]: ...



import logging
from importlib import import_module


def smart_load(conf):
    for key in conf:
        try:
            m = import_module(f'.{key}', package=__name__)
            if hasattr(m, 'load'):
                m.load(conf[key])
        except ModuleNotFoundError as err:
            logging.debug(f'load {key} failed: {err}')


__all__ = ('smart_load',)



import asyncio
import json
from io import BytesIO
from pathlib import Path

import aiofiles
import aiohttp


async def get_json(url, params, **kwargs):
    """
    GET method with json result
    Parameters
    ----------
    url : str
        API endpoint URL
    params : dict
        URL query parameters
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    dict
        Parsed JSON response
    """

    async with aiohttp.ClientSession() as session:
        async with session.get(url, params=params,
                               headers={'content-type': 'application/json'},
                               **kwargs) as resp:
            if resp.content_type == 'application/json':
                return await resp.json()
            else:
                return json.loads(await resp.text())


async def post_json(url, body, **kwargs):
    """
    post a json body with json result
    Parameters
    ----------
    url : str
        API endpoint URL
    body : dict
        Request body to be serialized as JSON
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    dict
        Parsed JSON response
    """
    async with aiohttp.ClientSession() as session:
        async with session.post(url, json=body, **kwargs) as resp:
            return await resp.json()


async def put_json(url, body, **kwargs):
    """
    PUT method with json request and response

    Parameters
    ----------
    url : str
        API endpoint URL
    body : dict
        Request body to be serialized as JSON
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    dict
        Parsed JSON response
    """
    async with aiohttp.ClientSession() as session:
        async with session.put(url, json=body, **kwargs) as resp:
            return await resp.json()


async def delete_json(url, **kwargs):
    """
    DELETE method with json response

    Parameters
    ----------
    url : str
        API endpoint URL
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    dict
        Parsed JSON response
    """
    async with aiohttp.ClientSession() as session:
        async with session.delete(url, **kwargs) as resp:
            return await resp.json()


async def download(url, **kwargs):
    """
    Download file from URL

    Parameters
    ----------
    url : str
        URL to download from
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    tuple
        BytesIO object with file content and content type
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(url, **kwargs) as resp:
            if resp.status >= 400:
                return None, None
            return BytesIO(await resp.content.read()), resp.content_type


async def download_to_file(url, destination_path, chunk_size=1024*1024, **kwargs):
    """
    Download file from URL and save to disk with progress tracking

    Parameters
    ----------
    url : str
        URL to download from
    destination_path : str or Path
        Path where to save the file
    chunk_size : int
        Size of chunks to download at once
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    bool
        True if download was successful, False otherwise
    """
    dest_path = Path(destination_path)
    if dest_path.exists() and not kwargs.get('overwrite', False):
        return False

    async with aiohttp.ClientSession() as session:
        async with session.get(url, **kwargs) as resp:
            if resp.status >= 400:
                return False

            # Ensure directory exists
            dest_path.parent.mkdir(parents=True, exist_ok=True)

            int(resp.headers.get('Content-Length', 0))
            downloaded = 0

            async with aiofiles.open(dest_path, 'wb') as f:
                async for chunk in resp.content.iter_chunked(chunk_size):
                    await f.write(chunk)
                    downloaded += len(chunk)

            return True


async def upload_file(url, file_path, field_name='file', additional_data=None, **kwargs):
    """
    Upload file to server using multipart/form-data

    Parameters
    ----------
    url : str
        Upload endpoint URL
    file_path : str or Path
        Path to the file to be uploaded
    field_name : str
        Form field name for the file
    additional_data : dict
        Additional form fields to include
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    dict
        Parsed JSON response
    """
    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"File not found: {file_path}")

    data = aiohttp.FormData()
    data.add_field(field_name,
                   open(file_path, 'rb'),
                   filename=file_path.name,
                   content_type='application/octet-stream')

    if additional_data:
        for key, value in additional_data.items():
            data.add_field(key, str(value))

    async with aiohttp.ClientSession() as session:
        async with session.post(url, data=data, **kwargs) as resp:
            return await resp.json()


async def upload_bytes(url, file_bytes, filename, field_name='file', content_type='application/octet-stream', additional_data=None, **kwargs):
    """
    Upload in-memory bytes to server using multipart/form-data

    Parameters
    ----------
    url : str
        Upload endpoint URL
    file_bytes : bytes or BytesIO
        File content as bytes or BytesIO
    filename : str
        Name to give the file
    field_name : str
        Form field name for the file
    content_type : str
        MIME type of the file
    additional_data : dict
        Additional form fields to include
    kwargs : dict
        Additional parameters to pass to the request

    Returns
    -------
    dict
        Parsed JSON response
    """
    data = aiohttp.FormData()

    if isinstance(file_bytes, BytesIO):
        file_bytes = file_bytes.getvalue()

    data.add_field(field_name,
                   file_bytes,
                   filename=filename,
                   content_type=content_type)

    if additional_data:
        for key, value in additional_data.items():
            data.add_field(key, str(value))

    async with aiohttp.ClientSession() as session:
        async with session.post(url, data=data, **kwargs) as resp:
            return await resp.json()


class EventSourceClient:
    """Client for Server-Sent Events (EventSource)"""

    def __init__(self, url, headers=None, reconnection_time=3.0, session=None):
        """
        Initialize EventSource client

        Parameters
        ----------
        url : str
            EventSource endpoint URL
        headers : dict, optional
            HTTP headers to send
        reconnection_time : float
            Time in seconds to wait before reconnecting
        session : aiohttp.ClientSession, optional
            Session to use for connections
        """
        self.url = url
        self.headers = headers or {}
        self.headers.update({'Accept': 'text/event-stream'})
        self.reconnection_time = reconnection_time
        self._session = session
        self._should_close_session = session is None
        self._event_callbacks = {}
        self._running = False

    async def __aenter__(self):
        if self._session is None:
            self._session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._running = False
        if self._should_close_session and self._session:
            await self._session.close()

    def on(self, event_name, callback):
        """
        Register callback for specific event type

        Parameters
        ----------
        event_name : str
            Name of the event to listen for ('message' for default events)
        callback : callable
            Function to call when event is received
        """
        self._event_callbacks[event_name] = callback
        return self

    async def _process_events(self, response):
        """Process event stream from response"""
        event_name = 'message'
        data = []
        last_id = None

        # Process the EventSource stream
        async for line in response.content:
            line = line.decode('utf-8').strip()

            if not line:
                # Empty line means dispatch the event
                if data and event_name in self._event_callbacks:
                    event_data = ''.join(data)
                    callback = self._event_callbacks[event_name]
                    await callback(event_data, last_id)

                # Reset for next event
                event_name = 'message'
                data = []
                continue

            if line.startswith(':'):
                # Comment, ignore
                continue

            if ':' in line:
                field, value = line.split(':', 1)
                if value.startswith(' '):
                    value = value[1:]

                if field == 'event':
                    event_name = value
                elif field == 'data':
                    data.append(value)
                elif field == 'id':
                    last_id = value
                elif field == 'retry':
                    try:
                        self.reconnection_time = int(value) / 1000.0
                    except ValueError:
                        pass

    async def connect(self):
        """
        Connect to EventSource endpoint and start processing events
        """
        self._running = True

        if self._session is None:
            self._session = aiohttp.ClientSession()
            self._should_close_session = True

        while self._running:
            try:
                async with self._session.get(self.url, headers=self.headers) as response:
                    if response.status != 200:
                        raise ConnectionError(f"Failed to connect to EventSource: {response.status}")

                    await self._process_events(response)
            except (aiohttp.ClientError, ConnectionError):
                if not self._running:
                    break
                await asyncio.sleep(self.reconnection_time)

    async def close(self):
        """Close the connection"""
        self._running = False
        if self._should_close_session and self._session:
            await self._session.close()
            self._session = None


class WebSocketClient:
    """Client for WebSocket connections"""

    def __init__(self, url, headers=None, session=None):
        """
        Initialize WebSocket client

        Parameters
        ----------
        url : str
            WebSocket endpoint URL
        headers : dict, optional
            HTTP headers for the initial connection
        session : aiohttp.ClientSession, optional
            Session to use for connections
        """
        self.url = url
        self.headers = headers or {}
        self._session = session
        self._should_close_session = session is None
        self._ws = None
        self._callbacks = {
            'message': [],
            'connect': [],
            'disconnect': [],
            'error': []
        }

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def on(self, event_type, callback):
        """
        Register callback for specific event

        Parameters
        ----------
        event_type : str
            Type of event: 'message', 'connect', 'disconnect', 'error'
        callback : callable
            Function to call when event occurs
        """
        if event_type in self._callbacks:
            self._callbacks[event_type].append(callback)
        return self

    async def _trigger_callbacks(self, event_type, *args):
        """Trigger all callbacks for an event type"""
        for callback in self._callbacks.get(event_type, []):
            if asyncio.iscoroutinefunction(callback):
                await callback(*args)
            else:
                callback(*args)

    async def connect(self):
        """Connect to WebSocket endpoint"""
        if self._session is None:
            self._session = aiohttp.ClientSession()
            self._should_close_session = True

        try:
            self._ws = await self._session.ws_connect(self.url, headers=self.headers)
            await self._trigger_callbacks('connect')
        except Exception as e:
            await self._trigger_callbacks('error', e)
            raise

    async def send_text(self, message):
        """
        Send text message to WebSocket

        Parameters
        ----------
        message : str
            Text message to send
        """
        if not self._ws:
            raise ConnectionError("WebSocket not connected")
        await self._ws.send_str(message)

    async def send_json(self, data):
        """
        Send JSON message to WebSocket

        Parameters
        ----------
        data : dict
            Data to be serialized as JSON and sent
        """
        if not self._ws:
            raise ConnectionError("WebSocket not connected")
        await self._ws.send_json(data)

    async def send_bytes(self, data):
        """
        Send binary message to WebSocket

        Parameters
        ----------
        data : bytes
            Binary data to send
        """
        if not self._ws:
            raise ConnectionError("WebSocket not connected")
        await self._ws.send_bytes(data)

    async def receive(self):
        """
        Receive a single message from WebSocket

        Returns
        -------
        aiohttp.WSMessage
            Received message
        """
        if not self._ws:
            raise ConnectionError("WebSocket not connected")
        msg = await self._ws.receive()

        if msg.type == aiohttp.WSMsgType.TEXT:
            await self._trigger_callbacks('message', msg.data, 'text')
        elif msg.type == aiohttp.WSMsgType.BINARY:
            await self._trigger_callbacks('message', msg.data, 'binary')
        elif msg.type == aiohttp.WSMsgType.ERROR:
            await self._trigger_callbacks('error', msg.data)
        elif msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSING):
            await self._trigger_callbacks('disconnect')

        return msg

    async def listen(self):
        """
        Listen for messages until connection is closed
        """
        if not self._ws:
            raise ConnectionError("WebSocket not connected")

        try:
            async for msg in self._ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    await self._trigger_callbacks('message', msg.data, 'text')
                elif msg.type == aiohttp.WSMsgType.BINARY:
                    await self._trigger_callbacks('message', msg.data, 'binary')
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    await self._trigger_callbacks('error', msg.data)
                    break
                elif msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSING):
                    break
        finally:
            await self._trigger_callbacks('disconnect')

    async def close(self):
        """Close WebSocket connection"""
        if self._ws:
            await self._ws.close()
            self._ws = None

        if self._should_close_session and self._session:
            await self._session.close()
            self._session = None



from jcutil.dba.async_ import read_sql, read_table
from jcutil.dba.helper import quick_update
from jcutil.dba.sqls import where

__all__ = [
    'read_sql',
    'read_table',
    'quick_update',
    'where'
]



from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

import consul as py_consul
import hcl
import yaml
from jcramda import decode, identity


def _hcl_load(raw_value):
    return hcl.loads(raw_value)


__all__ = (
    'Consul',
    'ConsulClient',
    'path_join',
    'fetch_key',
    'register_service',
    'deregister',
    'get_services',
    'find_service',
    'register_check',
    'deregister_check',
    'create_session',
    'destroy_session',
    'renew_session',
    'acquire_lock',
    'release_lock',
    'list_keys',
)

Consul = py_consul.Consul


class ConsulClient:
    """Consul客户端，提供同步API封装"""

    def __init__(self, host='127.0.0.1', port=8500, token=None, scheme='http', consistency='default', dc=None, verify=True):
        """初始化Consul客户端

        Args:
            host: Consul主机地址
            port: Consul端口
            token: ACL token
            scheme: 协议(http/https)
            consistency: 一致性模式
            dc: 数据中心
            verify: SSL验证
        """
        self.params = {
            'host': host,
            'port': port,
            'token': token,
            'scheme': scheme,
            'consistency': consistency,
            'dc': dc,
            'verify': verify
        }
        self._client = Consul(**self.params)

    @property
    def client(self) -> py_consul.Consul:
        """获取客户端"""
        return self._client

    def kv_get(self, key: str, **kwargs) -> Any:
        """获取键值"""
        return self._client.kv.get(key, **kwargs)

    def kv_put(self, key: str, value: str, **kwargs) -> bool:
        """设置键值"""
        return self._client.kv.put(key, value, **kwargs)

    def kv_list(self, prefix: str, **kwargs) -> Tuple[int, List[Dict]]:
        """列出指定前缀下的所有键"""
        return self._client.kv.get(prefix, recurse=True, **kwargs)

    def kv_delete(self, key: str, **kwargs) -> bool:
        """删除键值"""
        return self._client.kv.delete(key, **kwargs)

    def service_register(self, name: str, **kwargs) -> None:
        """注册服务"""
        self._client.agent.service.register(name, **kwargs)

    def service_deregister(self, service_id: str) -> None:
        """注销服务"""
        self._client.agent.service.deregister(service_id)

    def services(self) -> Dict:
        """获取所有服务"""
        return self._client.agent.services()

    def check_register(self, name: str, check: Dict, **kwargs) -> None:
        """注册健康检查"""
        self._client.agent.check.register(name, check=check, **kwargs)

    def check_deregister(self, check_id: str) -> None:
        """注销健康检查"""
        self._client.agent.check.deregister(check_id)

    def checks(self) -> Dict:
        """获取所有健康检查"""
        return self._client.agent.checks()

    def session_create(self, name: str = None, **kwargs) -> str:
        """创建会话

        Args:
            name: 会话名称
            **kwargs: 其他参数，如ttl、lock_delay等

        Returns:
            会话ID
        """
        session_id = self._client.session.create(name=name, **kwargs)
        return session_id

    def session_destroy(self, session_id: str) -> bool:
        """销毁会话

        Args:
            session_id: 会话ID

        Returns:
            是否成功
        """
        return self._client.session.destroy(session_id)

    def session_renew(self, session_id: str) -> bool:
        """续约会话

        Args:
            session_id: 会话ID

        Returns:
            是否成功
        """
        try:
            self._client.session.renew(session_id)
            return True
        except Exception:
            return False

    def lock_acquire(self, key: str, session_id: str, value: str = None) -> bool:
        """获取分布式锁

        Args:
            key: 锁键值
            session_id: 会话ID
            value: 锁的值，默认为None

        Returns:
            是否成功获取锁
        """
        return self._client.kv.put(key, value, acquire=session_id)

    def lock_release(self, key: str, session_id: str) -> bool:
        """释放分布式锁

        Args:
            key: 锁键值
            session_id: 会话ID

        Returns:
            是否成功释放锁
        """
        return self._client.kv.put(key, None, release=session_id)


# 全局客户端实例
_default_client = ConsulClient()


def path_join(*args):
    return '/'.join(args)


def _yaml_load(raw_value):
    return yaml.safe_load(raw_value)


def _json_load(raw_value):
    import json
    return json.loads(raw_value)


class ConfigFormat(Enum):
    Text = decode
    Number = Decimal
    Int = int
    Float = float
    Json = _json_load
    Yaml = _yaml_load
    Hcl = _hcl_load


def fetch_key(key_path, fmt: Callable = None, client: Optional[ConsulClient] = None) -> Any:
    """获取配置键值

    Args:
        key_path: 键路径
        fmt: 格式化函数
        client: Consul客户端实例，默认使用全局实例

    Returns:
        解析后的值
    """
    client = client or _default_client
    __, raw = client.kv_get(key_path)
    assert raw, f'not found any content in {key_path}'
    # noinspection PyCallingNonCallable
    values = raw.get('Value')
    return fmt(values) if callable(fmt) else values.decode()


def register_service(service_name, **kwargs):
    """注册服务

    Args:
        service_name: 服务名称
        **kwargs: 其他参数

    See Also:
        consul.base.Service
    """
    _default_client.service_register(service_name, **kwargs)


def deregister(service_id):
    """注销服务

    Args:
        service_id: 服务ID
    """
    _default_client.service_deregister(service_id)


def get_services(client: Optional[ConsulClient] = None) -> Dict:
    """获取所有已注册的服务

    Args:
        client: Consul客户端实例，默认使用全局实例

    Returns:
        包含所有已注册服务的字典，格式为 {service_id: service_info}
    """
    client = client or _default_client
    return client.services()


def find_service(query: str, by_id: bool = False, client: Optional[ConsulClient] = None) -> Dict:
    """按名称或ID查找服务

    Args:
        query: 服务名称或ID
        by_id: 是否按ID查找，默认为False（按名称查找）
        client: Consul客户端实例，默认使用全局实例

    Returns:
        匹配的服务字典，如果未找到则返回空字典
    """
    client = client or _default_client
    services = client.services()

    if by_id:
        # 直接按ID查找
        return {k: v for k, v in services.items() if k == query}
    else:
        # 按服务名称查找
        return {k: v for k, v in services.items() if v.get('Service') == query}


def register_check(name: str, check: Dict, **kwargs) -> None:
    """注册健康检查

    Args:
        name: 检查名称
        check: 检查配置
        **kwargs: 其他参数

    See Also:
        consul.agent.check.register
    """
    _default_client.check_register(name, check, **kwargs)


def deregister_check(check_id: str) -> None:
    """注销健康检查

    Args:
        check_id: 检查ID
    """
    _default_client.check_deregister(check_id)


def create_session(name: str = None, ttl: str = '30s', **kwargs) -> str:
    """创建Consul会话

    Args:
        name: 会话名称
        ttl: 会话超时时间
        **kwargs: 其他参数

    Returns:
        会话ID
    """
    return _default_client.session_create(name=name, ttl=ttl, **kwargs)


def destroy_session(session_id: str) -> bool:
    """销毁Consul会话

    Args:
        session_id: 会话ID

    Returns:
        是否成功
    """
    return _default_client.session_destroy(session_id)


def renew_session(session_id: str) -> bool:
    """续约Consul会话

    Args:
        session_id: 会话ID

    Returns:
        是否成功
    """
    return _default_client.session_renew(session_id)


def acquire_lock(key: str, session_id: str, value: str = None) -> bool:
    """获取分布式锁

    Args:
        key: 锁键值
        session_id: 会话ID
        value: 锁的值，默认为None

    Returns:
        是否成功获取锁
    """
    return _default_client.lock_acquire(key, session_id, value)


def release_lock(key: str, session_id: str) -> bool:
    """释放分布式锁

    Args:
        key: 锁键值
        session_id: 会话ID

    Returns:
        是否成功释放锁
    """
    return _default_client.lock_release(key, session_id)


def list_keys(prefix: str, client: Optional[ConsulClient] = None) -> List[Dict]:
    """列出指定前缀下的所有键

    Args:
        prefix: 键前缀
        client: Consul客户端实例，默认使用全局实例

    Returns:
        键值列表
    """
    client = client or _default_client
    index, data = client.kv_list(prefix)
    return data or []


class KvProperty:
    """Consul键值属性描述符"""

    def __init__(self, key, /, prefix=None, namespace=None, format=None, cached=None):
        self.key = key
        self._prefix = '/'.join(filter(None, (namespace or 'properties', prefix)))
        self._fmt = format or ConfigFormat.Text
        self._cached = cached

    def __get__(self, instance, cls):
        if instance is None:
            print(cls)
            return cls
        if callable(self.key):
            name = self.key.__name__
            func = self.key
        else:
            name = self.key
            func = identity
        value = func(fetch_key('/'.join([self._prefix, instance.__class__.__name__, name]), self._fmt))
        if self._cached:
            setattr(instance, name, value)
        return value



from enum import Enum
from typing import Any, Callable, Dict, List, NoReturn, Optional, Tuple

import consul  # 明确导入consul模块

class ConfigFormat(Enum):
    Text: Callable
    Number: Callable
    Int: Callable
    Float: Callable
    Json: Callable
    Yaml: Callable
    Hcl: Callable

def path_join(*args: str) -> str: ...

def fetch_key(key_path: str, fmt: Callable = None, client: Optional['ConsulClient'] = None) -> Any: ...

def register_service(service_name: str, **kwargs) -> NoReturn: ...

def deregister(service_id) -> NoReturn: ...

def get_services(client: Optional['ConsulClient'] = None) -> Dict: ...

def find_service(query: str, by_id: bool = False, client: Optional['ConsulClient'] = None) -> Dict: ...

def register_check(name: str, check: Dict, **kwargs) -> None: ...

def deregister_check(check_id: str) -> None: ...

def create_session(name: str = None, ttl: str = None, **kwargs) -> str: ...

def destroy_session(session_id: str) -> bool: ...

def renew_session(session_id: str) -> bool: ...

def acquire_lock(key: str, session_id: str, value: str = None) -> bool: ...

def release_lock(key: str, session_id: str) -> bool: ...

def list_keys(prefix: str, client: Optional['ConsulClient'] = None) -> List[Dict]: ...

class KvProperty:
  def __init__(self, key: str, /, prefix: str = None, namespace: str = None, format: callable = None, cached: bool = None) -> None: ...

class ConsulClient:
  def __init__(self, host: str = '127.0.0.1', port: int = 8500, token: str = None,
               scheme: str = 'http', consistency: str = 'default',
               dc: str = None, verify: bool = True) -> None: ...
  @property
  def client(self) -> consul.Consul: ...
  def kv_get(self, key: str, **kwargs) -> Any: ...
  def kv_put(self, key: str, value: str, **kwargs) -> bool: ...
  def kv_list(self, prefix: str, **kwargs) -> Tuple[int, List[Dict]]: ...
  def kv_delete(self, key: str, **kwargs) -> bool: ...
  def service_register(self, name: str, **kwargs) -> None: ...
  def service_deregister(self, service_id: str) -> None: ...
  def services(self) -> Dict: ...
  def check_register(self, name: str, check: Dict, **kwargs) -> None: ...
  def check_deregister(self, check_id: str) -> None: ...
  def checks(self) -> Dict: ...
  def session_create(self, name: str = None, **kwargs) -> str: ...
  def session_destroy(self, session_id: str) -> bool: ...
  def session_renew(self, session_id: str) -> bool: ...
  def lock_acquire(self, key: str, session_id: str, value: str = None) -> bool: ...
  def lock_release(self, key: str, session_id: str) -> bool: ...



import asyncio
import base64
import hmac
import os
from concurrent.futures.thread import ThreadPoolExecutor
from contextvars import copy_context
from functools import partial
from importlib import import_module

from jcramda import compose, curry, decode, encode, flatten, has_attr

from .jsonfy import (
    SafeJsonDecoder,
    SafeJsonEncoder,
    fix_document,
    from_json_file,
    pp_json,
    to_json,
    to_json_file,
    to_obj,
)
from .pdtools import TYPE_REGS, col_value, df_dt, df_to_dict, df_to_json, ser_to_json

__all__ = (
    'SafeJsonEncoder', 'SafeJsonDecoder', 'to_json', 'to_json_file',
    'pp_json', 'fix_document', 'to_obj', 'from_json_file',
    'col_value', 'df_dt', 'df_to_json', 'ser_to_json', 'df_to_dict', 'TYPE_REGS',
    'host_mac',
    'hmac_sha256',
    'uri_encode',
    'uri_decode',
    'nl_print',
    'c_write',
    'clear',
    'async_run',
    'load_fc',
    'obj_dumps',
    'obj_loads',
    'init_event_loop',
    'get_running_loop',
    'map_async',
    'utcnow',
)


def init_event_loop():
    try:
        # In Python 3.13+, get_event_loop() is deprecated
        # Use get_running_loop() or new_event_loop() instead
        try:
            # Try to get the running loop first
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # If no running loop, create a new one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
    except RuntimeError:
        # Fallback for older Python versions
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def get_running_loop():
    """
    Get the current running loop or create and set a new one if none exists.
    This function is designed to be safely used in async test environments,
    as it doesn't close the loop.
    """
    try:
        loop = asyncio.get_running_loop()
        # Check if the loop is closed
        if loop.is_closed():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
    except RuntimeError:
        # No running loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def host_mac():
    import uuid
    return hex(uuid.getnode())[2:].upper()


@curry
def hmac_sha256(key, s):
    cipher = hmac.new(key, s, digestmod='SHA256')
    return base64.b64encode(cipher.digest())


uri_encode = compose(decode, base64.urlsafe_b64decode, encode)
uri_decode = compose(decode, base64.urlsafe_b64decode)


# console cmd
nl_print = partial(print, end='\n\n')  # 多空一行的输出
c_write = partial(print, flush=False, end='')  # 不立刻输出的print
clear = partial(os.system, 'clear')


# async run
async def async_run(sync_func, *args, with_context=False, **kwargs):
    """
    Run a synchronous function in a thread executor asynchronously.

    Parameters
    ----------
    with_context : bool
        Whether to copy the current process context, default: False
    sync_func : callable
        The synchronous function to run
    args : tuple
        Positional arguments to pass to the function
    kwargs : dict
        Keyword arguments to pass to the function

    Returns
    -------
    Any
        The result of the function call
    """
    loop = get_running_loop()
    fn = partial(sync_func, *args, **kwargs)
    if with_context:
        fn = partial(copy_context().run, fn)
    try:
        return await loop.run_in_executor(None, fn)
    except RuntimeError as e:
        # If event loop is closed, get a new one and try again
        if "Event loop is closed" in str(e):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return await loop.run_in_executor(None, fn)
        raise


def load_fc(fc_name, module_name=None):
    package = None
    if ':' in fc_name:
        package = module_name
        module_name, fc_name = fc_name.rsplit(':', 1)
    assert module_name, 'module_name is not empty'
    if package and not module_name.startswith('.'):
        module_name = '.' + module_name
    w = import_module(module_name, package=package)
    return getattr(w, fc_name) if has_attr(fc_name)(w) else None


def obj_dumps(obj):
    from base64 import b64encode
    from pickle import dumps
    return b64encode(dumps(obj))


def obj_loads(raw):
    import base64
    import pickle
    return pickle.loads(base64.b64decode(raw))


def _splitor(data, start, limit):
    return data[start:start+limit]


def map_async(func, data, limit=None, splitor=_splitor):
    if limit is None:
        cpu_count = os.cpu_count() or 1
        data_len = len(data)
        # If data is smaller than CPU count, use 1 as limit
        # to ensure we process at least one item per chunk
        limit = max(1, int(data_len / cpu_count))
    loop = get_running_loop()
    tasks = []
    start = 0
    block = splitor(data, start, limit)
    with ThreadPoolExecutor() as pool:
        while len(block) > 0:
            tasks.append(loop.run_in_executor(pool, lambda d: [func(x) for x in d], block))
            start += limit
            block = splitor(data, start, limit)

    result = loop.run_until_complete(asyncio.gather(*tasks))
    return flatten(result)


def utcnow():
    from datetime import datetime, timezone
    return datetime.now(timezone.utc)




import logging
from importlib import import_module
from typing import Union

from jcramda import loc

__all__ = [
    'connect',
    'conn',
    'init_engine',
]

__engines = dict()


def init_engine(tag: str, *args, create_engine=None, **kwargs):
    """Init a oracle engine

    :param tag: a flag with connection engine
    :param create_engine: a function create a database connection engine
    :param kwargs: database connection config

    **Options**
     * schema: default is 'oracle', can use ['oracle', 'mysql', 'sqlite', 'postgres', etc...]
     * user: database connection username
     * password: database connection password
     * dsn: database host and port dsn string
     * url: database connection url, like that: "oracle://username:password@10.0.0.5:1521/sid?encoding=utf-8

    :return:
    """
    schema = kwargs.get('schema', 'oracle')
    if create_engine is None:
        sqlmodule = import_module('sqlalchemy')
        url = '{schema}://{user}:{password}@{dsn}?encoding=utf-8'\
            .format(schema=schema, user=kwargs['user'], password=kwargs['password'],
                    dsn=kwargs['dsn']) if 'url' not in kwargs else kwargs.pop('url')
        current_engine = sqlmodule.create_engine(
            url, pool_size=10, encoding='utf-8', **kwargs)
    else:
        current_engine = create_engine(*args, **kwargs)
    __engines[tag] = current_engine
    return current_engine


def new_client(tag, *args, create_engine=None, **kwargs):
    if create_engine is None:
        try:
            module = import_module('sqlalchemy')
            create_engine = getattr(module, 'create_engine')
        except ModuleNotFoundError as err:
            logging.error(err)

    __engines[tag] = create_engine(*args, **kwargs)
    return __engines[tag]


def get_client(name: Union[str, int] = 0):
    if len(__engines) > 0:
        return loc(name, __engines)
    raise RuntimeError('no any host can connect.')


def connect(n: Union[str, int] = 0):
    return get_client(n).connect()


def load(conf: dict):
    """
    一次性读取配置文件，生成数据库链接
    配置文件格式：dict(dbname="{dburl}")
    ```
    {
      "db1": "mysql://username:password@127.0.0.1:3306/dbname?encoding=utf8mb",
      "myoracle": "oracle://...",
    }
    ```
    @param conf: Dict[str, str]
    """
    if conf and len(conf) > 0:
        for key in conf:
            try:
                init_engine(key, url=conf[key])
            except Exception as err:
                logging.warning(f'load database [{key}] failed: {err}')
            # print(f'database [{key}] connected')


conn = connect


def instances():
    return [*__engines.keys()]



#
# 缓存实用工具
# @author Jochen.He
#
import asyncio
import base64
import logging
import os
import pickle
import shutil
import warnings
from contextvars import copy_context
from datetime import timedelta
from functools import partial, wraps
from pathlib import Path
from typing import Callable, Union
from uuid import uuid4

from jcramda import (
    _,
    compose,
    hexdigest,
    identity,
    join,
    replace,
)
from jcramda.core.operator import default_to
from joblib import dump, memory

from .core import to_json
from .defines import CACHE_DEFAULT_DIR, DEFAULT_CACHE_TIME, Writable

_obj_encode: Callable = compose(
    base64.encodebytes,
    pickle.dumps
)

_obj_decode: Callable = compose(
    pickle.loads,
    base64.decodebytes
)

__all__ = [
    'mem_cache',
    'redis_cache',
    'persistence',
    'clear_mem',
]


def mem_cache(cache_dir: str = CACHE_DEFAULT_DIR):
    """

    Parameters
    ----------
    cache_dir: a cache temp dir path

    Returns
    -------
    a wrapper function that be cached result
    """
    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)
    mem = memory.Memory(cache_dir, verbose=0)
    return mem.cache


def clear_mem(path: str = '', cache_dir: str = CACHE_DEFAULT_DIR):
    if '.' in path and '/' not in path:
        path = path.replace('.', os.path.sep)
    location = os.path.join(cache_dir, 'joblib', path)
    shutil.rmtree(location, ignore_errors=True)
    # warnings.warn(f'delete {location}')


try:
    from .drivers import redis

    def redis_cache(expires=DEFAULT_CACHE_TIME, prefix=None, redis_connector=None, result_assert=None):
        """
        cache function result in redis.
        result cache will be override by same process call

        Notes
        ----------
        this function is not thead safe
        ** function result must is a type can be json.dumps and json.loads **

        Parameters
        ----------
        prefix: cache key prefix, default is "fc"
        expires: Union[int, timedelta],
            seconds
        redis_connector: function,
            void -> redis.Client | redis.Pool
        result_assert: typing.Callable[[], bool]
            if this function is given, then use it check cache result, if return false then update cache

        """

        def decorator(f):
            f_name = f.__module__ + '.' + f.__name__
            c_key = join(':', [prefix, 'fc', f_name, '#pid'])
            key_gen = compose(
                replace('#pid', _, c_key),
                default_to(os.getpid()),
                hexdigest('md5'),
                to_json
            )

            @wraps(f)
            async def redis_cache_f(*args, update_cache: bool = False, **kwargs):
                client = redis_connector() if redis_connector else redis.connect()
                cache_key = key_gen({'args': args, 'kwargs': kwargs})

                if not update_cache and await client.exists(cache_key):
                    try:
                        raw_data = await client.get(cache_key)
                        r = _obj_decode(raw_data)
                        if not callable(result_assert) or result_assert(r):
                            return r
                    except (TypeError, ValueError, pickle.UnpicklingError) as err:
                        logging.error(err)

                r = f(*args, **kwargs)
                while asyncio.iscoroutine(r):
                    r = await r
                if r is not None:
                    expires_seconds = int(expires.total_seconds()) if isinstance(expires, timedelta) else expires
                    await client.set(cache_key, _obj_encode(r), ex=expires_seconds)
                return r

            if asyncio.iscoroutinefunction(f):
                @wraps(f)
                async def async_cache_f(*args, update_cache: bool = False, **kwargs):
                    return await redis_cache_f(*args, update_cache=update_cache, **kwargs)
                return async_cache_f

            @wraps(f)
            def sync_wrapper(*args, update_cache: bool = False, **kwargs):
                return asyncio.run(redis_cache_f(*args, update_cache=update_cache, **kwargs))

            return sync_wrapper

        return decorator
except ModuleNotFoundError:
    warnings.warn('redis is not installed, [redis_cache] is not cached!!')

    def redis_cache(expires=DEFAULT_CACHE_TIME, prefix=None, redis_connector=None, result_assert=None):
        def decorator(f):
            return f
        return decorator


def _save_data(fs: Writable, tmp_file: Path, data):
    try:
        dump(data, tmp_file)
        with tmp_file.open('rb') as ft:
            fs.write(ft)
    except FileExistsError:
        pass
    finally:
        tmp_file.unlink()
        fs.close()


def persistence(fs: Union[Writable, Callable[[], Writable]], f):
    """
    非阻塞式持久化指定方法的执行结果
    :param fs:
    :param f:
    :return:
    :python_version: >= 3.8
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        r = f(*args, **kwargs)
        if r is not None:
            save_opts = (identity(fs), Path(os.path.join(CACHE_DEFAULT_DIR, uuid4().hex)), r)
            asyncio.get_event_loop().run_in_executor(None, copy_context().run, partial(_save_data, *save_opts))
        return r

    return wrapper




"""
A simple aes encrypt and decrypt module with cryptodome
Support multiple encryption algorithms including AES, RSA, hash functions and more
"""
# depend on pycryptodomex module
import base64
import binascii
import hashlib
import hmac
from enum import IntEnum
from typing import Dict, Optional, Tuple, Union

from Cryptodome.Cipher import AES, PKCS1_OAEP
from Cryptodome.Hash import HMAC, SHA1, SHA256
from Cryptodome.PublicKey import RSA
from Cryptodome.Random import get_random_bytes
from Cryptodome.Signature import pkcs1_15
from Cryptodome.Util.Padding import pad, unpad
from jcramda import co, curry, enum_value, first
from jcramda.base.text import b64_urlsafe_encode

BS = AES.block_size
DEFAULT_ITERATIONS = 10000  # 默认PBKDF2迭代次数


class Mode(IntEnum):
    """AES加密模式"""
    EAX = AES.MODE_EAX
    ECB = AES.MODE_ECB
    CBC = AES.MODE_CBC
    OCB = AES.MODE_OCB
    CFB = AES.MODE_CFB
    OFB = AES.MODE_OFB
    CTR = AES.MODE_CTR
    CCM = AES.MODE_CCM
    GCM = AES.MODE_GCM
    SIV = AES.MODE_SIV


class HashAlgorithm(IntEnum):
    """哈希算法类型"""
    MD5 = 0
    SHA1 = 1
    SHA256 = 2
    SHA512 = 3


# AES加密解密相关函数
@curry
def aes_encrypt(key, plain: str, /, mode=None) -> Tuple[str, str]:
    """AES加密函数

    Args:
        key: 密钥，可以是十六进制字符串或字节
        plain: 待加密的明文
        mode: 加密模式，默认为AES.MODE_EAX

    Returns:
        (密文的十六进制表示, nonce/iv的十六进制表示)
    """
    key = bytes.fromhex(key) if isinstance(key, str) else key
    cipher = AES.new(key, enum_value(mode) or AES.MODE_EAX)
    # add padding to plaintext
    raw = pad(plain.encode(), BS)
    if hasattr(cipher, 'encrypt_and_digest'):
        ciphertext, tag = cipher.encrypt_and_digest(raw)
        nonce_or_iv = getattr(cipher, 'nonce', tag if enum_value(mode) == AES.MODE_SIV else b'')
    else:
        ciphertext = cipher.encrypt(raw)
        nonce_or_iv = getattr(cipher, 'iv', getattr(cipher, 'nonce', b''))
    return tuple(it.hex().upper() for it in (ciphertext, nonce_or_iv))


@curry
def aes_decrypt(key, value: str, /, mode=None, nonce_or_iv: str = None) -> str:
    """AES解密函数

    Args:
        key: 密钥，可以是十六进制字符串或字节
        value: 待解密的密文（十六进制字符串）
        mode: 解密模式，默认为AES.MODE_EAX
        nonce_or_iv: nonce或iv值（十六进制字符串）

    Returns:
        解密后的明文
    """
    key = bytes.fromhex(key) if isinstance(key, str) else key
    args = [key, enum_value(mode) or AES.MODE_EAX]
    kws = {}
    if nonce_or_iv:
        if enum_value(mode) == AES.MODE_CTR:
            kws['nonce'] = bytes.fromhex(nonce_or_iv)
        elif enum_value(mode) != AES.MODE_SIV:
            args.append(bytes.fromhex(nonce_or_iv))
    cryptor = AES.new(*args, **kws)
    if enum_value(mode) == AES.MODE_SIV:
        ciphertext = cryptor.decrypt_and_verify(bytes.fromhex(value), bytes.fromhex(nonce_or_iv))
    else:
        ciphertext = cryptor.decrypt(bytes.fromhex(value))
    return unpad(ciphertext, BS).decode()


def get_sha1prng_key(key: str) -> str:
    """SHA1PRNG密钥生成

    使用SHA1PRNG算法生成与Java AES加密兼容的密钥

    Args:
        key: 原始密钥字符串

    Returns:
        生成的密钥（十六进制字符串）
    """
    return hashlib.sha1(
        hashlib.sha1(key.encode()).digest()
    ).hexdigest().upper()[:32]


# 预定义的AES加密解密函数
aes_ecb_encrypt = co(first, aes_encrypt(mode=AES.MODE_ECB))
aes_ecb_decrypt = aes_decrypt(mode=AES.MODE_ECB)
aes_cfb_encrypt = aes_encrypt(mode=AES.MODE_CFB)
aes_cfb_decrypt = aes_decrypt(mode=AES.MODE_CFB)
aes_cbc_encrypt = aes_encrypt(mode=AES.MODE_CBC)
aes_cbc_decrypt = aes_decrypt(mode=AES.MODE_CBC)
aes_eax_encrypt = aes_encrypt(mode=AES.MODE_EAX)
aes_eax_decrypt = aes_decrypt(mode=AES.MODE_EAX)


# PBKDF2密钥生成
def pbkdf2_key(password: str, salt: Optional[bytes] = None, iterations: int = DEFAULT_ITERATIONS, key_length: int = 32) -> Tuple[bytes, bytes]:
    """使用PBKDF2算法生成密钥

    Args:
        password: 密码明文
        salt: 盐值，如果为None则随机生成
        iterations: 迭代次数，默认10000
        key_length: 密钥长度，默认32字节

    Returns:
        (密钥, 盐值)
    """
    if salt is None:
        salt = get_random_bytes(16)
    derived_key = hashlib.pbkdf2_hmac('sha256', password.encode(), salt, iterations, key_length)
    return derived_key, salt


def generate_aes_key(length: int = 32) -> str:
    """生成随机AES密钥

    Args:
        length: 密钥长度（字节），默认32

    Returns:
        随机密钥（十六进制字符串）
    """
    return get_random_bytes(length).hex().upper()


# Base64编码解码函数
def to_base64(data: Union[str, bytes]) -> str:
    """转换为Base64编码

    Args:
        data: 待编码的数据，可以是字符串或字节

    Returns:
        Base64编码后的字符串
    """
    if isinstance(data, str):
        if all(c in '0123456789ABCDEFabcdef' for c in data) and len(data) % 2 == 0:
            # 如果是十六进制字符串，先转换为字节
            data = bytes.fromhex(data)
        else:
            data = data.encode()
    return base64.b64encode(data).decode()


def from_base64(data: str) -> bytes:
    """从Base64解码

    Args:
        data: Base64编码的字符串

    Returns:
        解码后的字节
    """
    return base64.b64decode(data)


def to_base64_url_safe(data: Union[str, bytes]) -> str:
    """转换为URL安全的Base64编码

    Args:
        data: 待编码的数据，可以是字符串或字节

    Returns:
        URL安全的Base64编码字符串
    """
    if isinstance(data, str):
        if all(c in '0123456789ABCDEFabcdef' for c in data) and len(data) % 2 == 0:
            data = bytes.fromhex(data)
        else:
            data = data.encode()
    return b64_urlsafe_encode(data)


# 哈希函数
def hash_sha1(data: Union[str, bytes]) -> str:
    """计算SHA1哈希值

    Args:
        data: 待哈希的数据，可以是字符串或字节

    Returns:
        十六进制哈希值（大写）
    """
    if isinstance(data, str):
        data = data.encode()
    return hashlib.sha1(data).hexdigest().upper()


def hash_sha256(data: Union[str, bytes]) -> str:
    """计算SHA256哈希值

    Args:
        data: 待哈希的数据，可以是字符串或字节

    Returns:
        十六进制哈希值（大写）
    """
    if isinstance(data, str):
        data = data.encode()
    return hashlib.sha256(data).hexdigest().upper()


def hash_sha512(data: Union[str, bytes]) -> str:
    """计算SHA512哈希值

    Args:
        data: 待哈希的数据，可以是字符串或字节

    Returns:
        十六进制哈希值（大写）
    """
    if isinstance(data, str):
        data = data.encode()
    return hashlib.sha512(data).hexdigest().upper()


def hash_md5(data: Union[str, bytes]) -> str:
    """计算MD5哈希值

    Args:
        data: 待哈希的数据，可以是字符串或字节

    Returns:
        十六进制哈希值（大写）
    """
    if isinstance(data, str):
        data = data.encode()
    return hashlib.md5(data).hexdigest().upper()


def sha3sum(plaintext: Union[str, bytes]) -> str:
    """计算SHA3-256哈希值

    Args:
        plaintext: 待哈希的数据，可以是字符串或字节

    Returns:
        十六进制哈希值（大写）
    """
    if isinstance(plaintext, str):
        plaintext = plaintext.encode()
    return hashlib.sha3_256(plaintext).hexdigest().upper()


# HMAC函数
def hmac_sha256(key: Union[str, bytes], message: Union[str, bytes]) -> str:
    """计算HMAC-SHA256值

    Args:
        key: 密钥，可以是字符串或字节
        message: 消息，可以是字符串或字节

    Returns:
        十六进制HMAC值（大写）
    """
    if isinstance(key, str):
        key = key.encode()
    if isinstance(message, str):
        message = message.encode()
    h = HMAC.new(key, message, SHA256)
    return h.hexdigest().upper()


def hmac_sha1(key: Union[str, bytes], message: Union[str, bytes]) -> str:
    """计算HMAC-SHA1值

    Args:
        key: 密钥，可以是字符串或字节
        message: 消息，可以是字符串或字节

    Returns:
        十六进制HMAC值（大写）
    """
    if isinstance(key, str):
        key = key.encode()
    if isinstance(message, str):
        message = message.encode()
    h = HMAC.new(key, message, SHA1)
    return h.hexdigest().upper()


# RSA加解密相关函数
def generate_rsa_key_pair(bits: int = 2048) -> Dict[str, bytes]:
    """生成RSA密钥对

    Args:
        bits: 密钥位数，默认2048

    Returns:
        包含私钥和公钥的字典
    """
    key = RSA.generate(bits)
    private_key = key.export_key()
    public_key = key.publickey().export_key()
    return {
        'private_key': private_key,
        'public_key': public_key
    }


def rsa_encrypt(public_key: Union[str, bytes], plain_data: Union[str, bytes]) -> str:
    """RSA加密

    Args:
        public_key: RSA公钥，可以是PEM格式的字符串或字节
        plain_data: 待加密的数据，可以是字符串或字节

    Returns:
        加密后的数据（Base64编码）
    """
    if isinstance(public_key, str):
        public_key = public_key.encode()
    if isinstance(plain_data, str):
        plain_data = plain_data.encode()

    key = RSA.import_key(public_key)
    cipher = PKCS1_OAEP.new(key)
    ciphertext = cipher.encrypt(plain_data)
    return base64.b64encode(ciphertext).decode()


def rsa_decrypt(private_key: Union[str, bytes], cipher_data: str) -> bytes:
    """RSA解密

    Args:
        private_key: RSA私钥，可以是PEM格式的字符串或字节
        cipher_data: 待解密的数据（Base64编码）

    Returns:
        解密后的原始数据
    """
    if isinstance(private_key, str):
        private_key = private_key.encode()

    key = RSA.import_key(private_key)
    cipher = PKCS1_OAEP.new(key)
    ciphertext = base64.b64decode(cipher_data)
    return cipher.decrypt(ciphertext)


def rsa_sign(private_key: Union[str, bytes], data: Union[str, bytes]) -> str:
    """RSA签名

    Args:
        private_key: RSA私钥，可以是PEM格式的字符串或字节
        data: 待签名的数据，可以是字符串或字节

    Returns:
        签名值（Base64编码）
    """
    if isinstance(private_key, str):
        private_key = private_key.encode()
    if isinstance(data, str):
        data = data.encode()

    key = RSA.import_key(private_key)
    h = SHA256.new(data)
    signature = pkcs1_15.new(key).sign(h)
    return base64.b64encode(signature).decode()


def rsa_verify(public_key: Union[str, bytes], data: Union[str, bytes], signature: str) -> bool:
    """验证RSA签名

    Args:
        public_key: RSA公钥，可以是PEM格式的字符串或字节
        data: 原始数据，可以是字符串或字节
        signature: 签名值（Base64编码）

    Returns:
        签名是否有效
    """
    if isinstance(public_key, str):
        public_key = public_key.encode()
    if isinstance(data, str):
        data = data.encode()

    key = RSA.import_key(public_key)
    h = SHA256.new(data)
    try:
        pkcs1_15.new(key).verify(h, base64.b64decode(signature))
        return True
    except (ValueError, TypeError):
        return False


# 实用工具函数
def secure_compare(a: Union[str, bytes], b: Union[str, bytes]) -> bool:
    """安全比较两个字符串或字节，防止时序攻击

    Args:
        a: 第一个字符串或字节
        b: 第二个字符串或字节

    Returns:
        两个值是否相等
    """
    if isinstance(a, str):
        a = a.encode()
    if isinstance(b, str):
        b = b.encode()

    return hmac.compare_digest(a, b)


def generate_random_string(length: int = 16) -> str:
    """生成随机字符串

    Args:
        length: 字符串长度，默认16

    Returns:
        随机字符串（十六进制）
    """
    return binascii.hexlify(get_random_bytes(length)).decode()



from enum import IntEnum
from functools import partial
from typing import Callable, List, NoReturn, Sized, Tuple

Color: IntEnum

FontFormat: IntEnum

EndFlag: IntEnum

class Chalk(Sized, Callable):
    def __len__(self) -> int: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __call__(self, *args, **kwargs) -> str: ...
    def __add__(self, other: Chalk) -> Chalk: ...

    __buffer__: List[str]
    __chains__: List[str]
    def __init__(self, text: str = ..., fgc: Color = ..., bgc: Color = ...,
                 styles: Tuple[FontFormat] = ...): ...
    def use(self, *args, **kwargs) -> Chalk: ...
    def text(self, text: str) -> Chalk: ...
    def format(self, text: str, *style: FontFormat) -> Chalk: ...
    def bold(self, text: str) -> Chalk: ...
    def end(self, *flag: EndFlag) -> Chalk: ...
    def expandtabs(self) -> str: ...
    @property
    def raw(self) -> str: ...


RedChalk: partial[Chalk]
GreenChalk: partial[Chalk]
BlueChalk: partial[Chalk]
YellowChalk: partial[Chalk]
MagentaChalk: partial[Chalk]
CyanChalk: partial[Chalk]
WhiteChalk: partial[Chalk]
BlackChalk: partial[Chalk]
BoldChalk: partial[Chalk]

BrightBlackChalk: partial[Chalk]
BrightBlueChalk = partial[Chalk]
BrightCyanChalk = partial[Chalk]
BrightGreenChalk = partial[Chalk]
BrightMagentaChalk = partial[Chalk]
BrightRedChalk = partial[Chalk]
BrightYellowChalk = partial[Chalk]
BrightWhiteChalk = partial[Chalk]


def show_menu(_items: list, is_submenu: bool = ..., title: str = ...) -> NoReturn: ...
def select() -> int: ...
